import sys
from functools import partial

import dask
import numpy as np
import pytest
from scipy import ndimage, stats

from raster_tools import Raster, focal
from tests.utils import assert_valid_raster

PY_VER_37 = sys.version_info[0] == 3 and sys.version_info[1] == 7


def test_get_focal_window_circle_rect():
    truths = [
        np.array([[1.0]]),
        np.array([[0.0, 1.0, 0.0], [1.0, 1.0, 1.0], [0.0, 1.0, 0.0]]),
        np.array(
            [
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 0.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
            ],
        ),
        np.array(
            [
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
            ],
        ),
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            ],
        ),
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
        ),
    ]
    truths = [t.astype(bool) for t in truths]
    for r, truth in zip(range(1, len(truths) + 1), truths):
        window = focal.get_focal_window(r)
        assert np.allclose(window, truth)
        assert window.dtype == bool
    for w in range(1, 6):
        for h in range(1, 6):
            window = focal.get_focal_window(w, h)
            assert np.allclose(window, np.ones((w, h)))
            assert window.dtype == bool


def test_get_focal_window_annulus():
    truths = [
        np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]]),
        np.array(
            [
                [0, 0, 1, 0, 0],
                [0, 1, 1, 1, 0],
                [1, 1, 0, 1, 1],
                [0, 1, 1, 1, 0],
                [0, 0, 1, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 1, 0, 0],
                [0, 1, 0, 1, 0],
                [1, 0, 0, 0, 1],
                [0, 1, 0, 1, 0],
                [0, 0, 1, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 1, 0, 0, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [1, 1, 1, 0, 1, 1, 1],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 1, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 1, 0, 0, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 0, 1, 1, 0],
                [1, 1, 0, 0, 0, 1, 1],
                [0, 1, 1, 0, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 1, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 1, 0, 0, 0],
                [0, 1, 1, 0, 1, 1, 0],
                [0, 1, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 1, 0],
                [0, 1, 1, 0, 1, 1, 0],
                [0, 0, 0, 1, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [1, 1, 1, 1, 0, 1, 1, 1, 1],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 0, 1, 1, 1, 0],
                [1, 1, 1, 0, 0, 0, 1, 1, 1],
                [0, 1, 1, 1, 0, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 0, 1, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 1, 1, 0],
                [1, 1, 0, 0, 0, 0, 0, 1, 1],
                [0, 1, 1, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 1, 0, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 1, 1, 0, 1, 1, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 1, 0],
                [0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 0, 0, 1, 0],
                [0, 1, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 1, 1, 0, 1, 1, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1],
                [0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        ),
        np.array(
            [
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 0],
                [0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            ]
        ),
    ]
    truths = [t.astype(bool) for t in truths]
    i = 0
    for r2 in range(1, 7):
        for r1 in range(1, r2):
            if r1 >= r2:
                continue
            w = focal.get_focal_window((r1, r2))
            t = truths[i]
            assert np.allclose(w, t)
            i += 1


def test_get_focal_window_errors():
    for r in [-2, -1, 0]:
        with pytest.raises(ValueError):
            focal.get_focal_window(r)
    for r in [2.3, 4.999, 3.0, None, "4"]:
        with pytest.raises(TypeError):
            focal.get_focal_window(r)
    for rvalues in [(0, 3), (3, 3), (3, 1), (-1, 3), (-3, -1)]:
        with pytest.raises(ValueError):
            focal.get_focal_window(rvalues)
    for rvalues in [(3.0, 4), (2.1, 3.0), (3, 5.0)]:
        with pytest.raises(TypeError):
            focal.get_focal_window(rvalues)
    for args in [(-1, 4), (0, 3), (3, -3)]:
        with pytest.raises(ValueError):
            focal.get_focal_window(*args)
    for args in [(3.0, 3.0), (4, 3.0)]:
        with pytest.raises(TypeError):
            focal.get_focal_window(*args)
    with pytest.raises(ValueError):
        focal.get_focal_window((2, 4), 5)


def test_focal_return_dask():
    x = np.arange(16.0).reshape(4, 4)
    kern = focal.get_focal_window(2)
    assert dask.is_dask_collection(focal._focal(x, kern, "max", True))
    assert dask.is_dask_collection(focal._focal(x, kern, "max", False))
    xd = dask.array.from_array(x)
    assert dask.is_dask_collection(focal._focal(xd, kern, "max", True))
    assert dask.is_dask_collection(focal._focal(xd, kern, "max", False))


def asm(x):
    c = {}
    n = 0
    for v in x.ravel():
        if not np.isnan(v):
            if v in c:
                c[v] += 1
            else:
                c[v] = 1
            n += 1
    if n == 0:
        return 0.0
    p = np.array([cnt / n for cnt in c.values()])
    return np.sum(p * p)


def entropy(x):
    c = {}
    xn = x[~np.isnan(x)]
    for v in xn:
        if v in c:
            c[v] += 1
        else:
            c[v] = 1
    n = sum(c.values())
    if not n:
        return 0.0
    p = np.array([cnt / n for cnt in c.values()])
    return np.nansum(-p * np.log(p))


def mode(x):
    if x[~np.isnan(x)].size == 0:
        return np.nan
    mode_kwargs = {"axis": None, "nan_policy": "omit"}
    if not PY_VER_37:
        mode_kwargs["keepdims"] = True
    m = stats.mode(x, **mode_kwargs)
    return m.mode[0]


def unique(x):
    u = np.unique(x[~np.isnan(x)])
    if len(u):
        return len(u)
    return 0


def correlate(x, kern):
    return np.nansum(x * kern.ravel())


@pytest.mark.parametrize(
    "kernel,kernel_params",
    [
        (focal.get_focal_window(2), (2,)),
        (focal.get_focal_window(3, 3), (3, 3)),
        (focal.get_focal_window((2, 3)), ((2, 3),)),
    ],
)
@pytest.mark.parametrize(
    "filter_name,filter_func",
    [
        ("asm", asm),
        ("entropy", entropy),
        ("max", np.nanmax),
        ("mean", np.nanmean),
        ("median", np.nanmedian),
        ("mode", mode),
        ("min", np.nanmin),
        ("std", np.nanstd),
        ("sum", np.nansum),
        ("var", np.nanvar),
        ("unique", unique),
    ],
)
@pytest.mark.filterwarnings("ignore:All-NaN slice encountered")
@pytest.mark.filterwarnings("ignore:Mean of empty slice")
@pytest.mark.filterwarnings("ignore:Degrees of freedom <= 0 for slice")
def test_focal(filter_name, filter_func, kernel, kernel_params):
    x = np.arange(64.0).reshape(8, 8)
    x[:3, :3] = np.nan
    mask = np.isnan(x)
    rx = Raster(x).set_null_value(np.nan)
    truth = ndimage.generic_filter(
        rx.xdata.data[0].compute(),
        filter_func,
        footprint=kernel,
        mode="constant",
        cval=np.nan,
    )
    res = focal._focal(x, kernel, filter_name, True).compute()
    assert np.allclose(truth, res, equal_nan=True)

    result_raster = focal.focal(rx, filter_name, *kernel_params)
    assert_valid_raster(result_raster)
    res_full = result_raster.values[0]
    # Fill with null value
    assert np.allclose(
        np.where(mask, result_raster.null_value, truth),
        res_full,
        equal_nan=True,
    )


def test_correlate_return_dask():
    x = np.arange(16.0).reshape(4, 4)
    kern = focal.get_focal_window(2)
    assert dask.is_dask_collection(focal._correlate(x, kern))


def test_focal_correlate():
    for kern in [np.ones((5, 5)), np.ones((4, 4))]:
        func = partial(correlate, kern=kern)
        for nan_aware in [False, True]:
            data = np.arange(64.0).reshape(8, 8)
            if nan_aware:
                data[:3, :3] = np.nan
            for mode in ["reflect", "nearest", "wrap", "constant"]:
                origin = [-1 if d % 2 == 0 else 0 for d in kern.shape]
                truth = ndimage.generic_filter(
                    data, func, size=kern.shape, mode=mode, origin=origin
                )
                test = focal._correlate(
                    data, kern, mode=mode, nan_aware=nan_aware
                ).compute()
                assert np.allclose(truth, test, equal_nan=nan_aware)


def test_focal_integration():
    rs = Raster("tests/data/raster/multiband_small.tif")
    rsnp = rs.values
    truth = rsnp.astype(float)
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd], np.nanmean, size=3, mode="constant", cval=np.nan
        )
    res_raster = focal.focal(rs, "mean", 3, 3)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    assert np.allclose(truth, res, equal_nan=True)
    truth = rsnp.astype(float)
    kern = focal.get_focal_window(3)
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd],
            np.nanmedian,
            footprint=kern,
            mode="constant",
            cval=np.nan,
        )
    res_raster = focal.focal(rs, "median", 3)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    assert np.allclose(truth, res, equal_nan=True)


def test_focal_integration_raster_input():
    rs = Raster("tests/data/raster/multiband_small.tif")
    rsnp = rs.values
    with pytest.raises(TypeError):
        focal.focal(rsnp, "median", 3)
    res_raster = focal.focal(rs, "mean", 1)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    assert np.allclose(rsnp, res, equal_nan=True)


def test_focal_output_type():
    rs = Raster("tests/data/raster/multiband_small.tif") * 100
    rs_masked = rs.set_null_value(-1).astype(int)
    rsi = rs.set_null_value(None).astype(int)

    # Masked
    assert rs_masked._masked
    assert rs_masked.dtype.kind == "i"
    res = focal.focal(rs_masked, "mode", 3)
    assert_valid_raster(res)
    res = res.eval()
    assert res.dtype.kind == "f"
    res = focal.focal(rs_masked, "unique", 3)
    assert_valid_raster(res)
    res = res.eval()
    assert res.dtype.kind == "f"
    res = focal.focal(rs_masked, "mean", 3)
    assert_valid_raster(res)
    res = res.eval()
    assert res.dtype.kind == "f"

    # Unmasked
    assert not rsi._masked
    assert rsi.dtype.kind == "i"
    res = focal.focal(rsi, "mode", 3).eval()
    assert res.dtype.kind == "i"
    res = focal.focal(rsi, "unique", 3).eval()
    assert res.dtype.kind == "u"
    res = focal.focal(rsi, "mean", 3).eval()
    assert res.dtype.kind == "f"


def test_correlate_integration():
    rs = Raster("tests/data/raster/multiband_small.tif").astype(float)
    rsnp = rs.values
    truth = rsnp.astype(float)
    kernel = np.array([[1, 1, 1], [1, 1, 0], [1, 0, 0]]).astype(float)
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd], np.sum, footprint=kernel, mode="constant"
        )
    res_raster = focal.correlate(rs, kernel)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    assert np.allclose(truth, res, equal_nan=False)

    truth = rsnp.astype(float)
    truth[:, :3, :3] = np.nan
    kern = focal.get_focal_window(3)
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd],
            np.nansum,
            footprint=kern,
            mode="constant",
        )
    rs.xdata.data[:, :3, :3] = -1
    rs = rs.set_null_value(-1)
    res_raster = focal.correlate(rs, kern)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    truth[res_raster.xmask.values] = res_raster.null_value
    assert np.allclose(truth, res, equal_nan=True)


def test_convolve_integration():
    rs = Raster("tests/data/raster/multiband_small.tif").astype(float)
    rsnp = rs.values
    truth = rsnp.astype(float)
    kernel = np.array([[1, 1, 1], [1, 1, 0], [1, 0, 0]]).astype(float)
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd],
            np.sum,
            footprint=kernel[::-1, ::-1],
            mode="constant",
        )
    res_raster = focal.convolve(rs, kernel)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    assert np.allclose(truth, res, equal_nan=False)

    truth = rsnp.astype(float)
    truth[:, :3, :3] = np.nan
    for bnd in range(truth.shape[0]):
        truth[bnd] = ndimage.generic_filter(
            truth[bnd],
            np.nansum,
            footprint=kernel[::-1, ::-1],
            mode="constant",
        )
    rs.xdata.data[:, :3, :3] = -1
    rs = rs.set_null_value(-1)
    res_raster = focal.convolve(rs, kernel)
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster.values
    truth[res_raster.xmask.values] = res_raster.null_value
    assert np.allclose(truth, res, equal_nan=True)


def test_correlate_integration_raster_input():
    rs = Raster("tests/data/raster/multiband_small.tif")
    rsnp = rs.values
    with pytest.raises(TypeError):
        focal.correlate(rsnp, 3)
    res_raster = focal.correlate(rs, np.ones((1, 1)))
    assert_valid_raster(res_raster)
    assert res_raster.crs == rs.crs
    res = res_raster
    assert np.allclose(rsnp, res, equal_nan=True)


def test_correlate_output_type():
    rs = Raster("tests/data/raster/multiband_small.tif") * 100
    rs = rs.set_null_value(-1)
    rs = rs.astype(int)

    assert rs._masked
    assert rs.dtype.kind == "i"
    res = focal.correlate(rs, np.ones((3, 3), dtype=int)).eval()
    assert res.dtype == rs.dtype

    res = focal.correlate(rs, np.ones((3, 3), dtype=float)).eval()
    assert res.dtype.kind == "f"



from collections.abc import Iterable, Iterator, Sequence
from functools import partial

import dask
import dask.array as da
import dask.dataframe as dd
import numba as nb
import numpy as np
import pandas as pd
from dask_image import ndmeasure

from raster_tools.dask_utils import dask_nanmax, dask_nanmin
from raster_tools.dtypes import F64, I64, is_int, is_str
from raster_tools.raster import Raster, get_raster, xy_to_rowcol
from raster_tools.vector import Vector, get_vector

__all__ = ["ZONAL_STAT_FUNCS", "zonal_stats"]


def _nan_count(x):
    return da.count_nonzero(~np.isnan(x))


def _nan_median(x):
    x = da.asarray(x)
    return da.nanmedian(x, axis=0)


def _nan_unique(x):
    return _nan_count(da.unique(x))


def _flatten_gen(x):
    """
    A generator that recursively yields numpy arrays from arbitrarily nested
    lists of arrays.
    """
    for xi in x:
        if isinstance(x, Iterable) and not isinstance(xi, np.ndarray):
            yield from _flatten_gen(xi)
        else:
            yield xi


def _flatten(x):
    """Flatten nested lists of arrays."""
    if isinstance(x, np.ndarray):
        return [x]
    return list(_flatten_gen(x))


def _recursive_map(func, *seqs):
    """Apply a function to items in nested sequences."""
    if isinstance(seqs[0], (list, Iterator)):
        return [_recursive_map(func, *items) for items in zip(*seqs)]
    return func(*seqs)


def _unique_with_counts_chunk(x, computing_meta=False, axis=(), **kwargs):
    """Reduce a dask chunk to a dict of unique values and counts.

    This is the leaf operation in the reduction tree.
    """
    if computing_meta:
        return x
    x_non_nan = x[~np.isnan(x)]
    values, counts = np.unique(x_non_nan, return_counts=True)
    while values.ndim < len(axis):
        values = np.expand_dims(values, axis=0)
        counts = np.expand_dims(counts, axis=0)
    return {"values": values, "counts": counts}


def _ravel_key(item, key):
    return item[key].ravel()


_ravel_values = partial(_ravel_key, key="values")
_ravel_counts = partial(_ravel_key, key="counts")


def _split_concat(pairs, split_func):
    # Split out a key from lists of dicts, ravel them, and concat all together
    split = _recursive_map(split_func, pairs)
    return np.concatenate(_flatten(split))


def _unique_with_counts_combine(
    pairs, computing_meta=False, axis=(), **kwargs
):
    """Merge/combine branches of the unique-with-counts reduction tree.

    This includes results from multiple _unique_with_counts_chunk calls and
    from prior _unique_with_counts_combine calls.
    """
    values = (
        _recursive_map(_ravel_values, pairs) if not computing_meta else pairs
    )
    values = np.concatenate(_flatten(values))
    if computing_meta:
        return np.array([[[0]]], dtype=pairs.dtype)

    counts = _split_concat(pairs, _ravel_counts)
    res = {v: 0 for v in values}
    for v, c in zip(values, counts):
        res[v] += c
    values = np.array(list(res.keys()))
    counts = np.array(list(res.values()))
    while values.ndim < len(axis):
        values = np.expand_dims(values, axis=0)
        counts = np.expand_dims(counts, axis=0)
    return {"values": values, "counts": counts}


def _mode_agg(pairs, computing_meta=False, axis=(), **kwargs):
    """Perform the final aggregation to a single mode value."""
    values = (
        _split_concat(pairs, _ravel_values) if not computing_meta else pairs
    )
    if computing_meta:
        return pairs.dtype.type(0)
    if len(values) == 0:
        # See note below about wrapping in np.array()
        return np.array(np.nan)
    counts = _split_concat(pairs, _ravel_counts)
    res = {v: 0 for v in values}
    for v, c in zip(values, counts):
        res[v] += c
    values = res.keys()
    counts = res.values()
    sorted_pairs = sorted(zip(counts, values), reverse=True)
    # Find the minimum mode when there is a tie. This is the same behavior as
    # scipy.
    i = -1
    c = sorted_pairs[0][0]
    for pair in sorted_pairs:
        if pair[0] == c:
            i += 1
        else:
            break
    # NOTE: wrapping the value in an array is a hack to prevent dask from
    # mishandling the return value as an array with dims, leading to index
    # errors. I can't pierce the veil of black magic that is causing the
    # mishandling so this is the best fix I can come up with.
    return np.array(sorted_pairs[i][1])


def _nan_mode(x):
    """
    Compute the statistical mode of an array using a dask reduction operation.
    """
    return da.reduction(
        x,
        chunk=_unique_with_counts_chunk,
        combine=_unique_with_counts_combine,
        aggregate=_mode_agg,
        # F64 to allow for potential empty input array. In that case a NaN is
        # returned.
        dtype=F64,
        # Turn off concatenation to prevent dask from trying to concat the
        # dicts of variable length values and counts. Dask tries to concat
        # along the wrong axis, which causes errors.
        concatenate=False,
    )


@nb.jit(nopython=True, nogil=True)
def _entropy(values, counts):
    if len(values) == 0:
        return np.nan
    res = {v: 0 for v in values}
    for v, c in zip(values, counts):
        res[v] += c
    counts = res.values()
    entropy = 0.0
    frac = 1 / len(res)
    for cnt in counts:
        p = cnt * frac
        entropy -= p * np.log(p)
    return entropy


@nb.jit(nopython=True, nogil=True)
def _asm(values, counts):
    if len(values) == 0:
        return np.nan
    res = {v: 0 for v in values}
    for v, c in zip(values, counts):
        res[v] += c
    counts = res.values()
    asm = 0.0
    frac = 1 / len(res)
    for cnt in counts:
        p = cnt * frac
        asm += p * p
    return asm


def _entropy_asm_agg(
    pairs, compute_entropy, computing_meta=False, axis=(), **kwargs
):
    """Perform the final aggregation to a single entropy or ASM value."""
    if computing_meta:
        return 0
    values = _split_concat(pairs, _ravel_values)
    if len(values) == 0:
        return np.array([])
    counts = _split_concat(pairs, _ravel_counts)
    # NOTE: wrapping the value in an array is a hack to prevent dask from
    # mishandling the return value as an array with dims, leading to index
    # errors. I can't pierce the veil of black magic that is causing the
    # mishandling so this is the best fix I can come up with.
    if compute_entropy:
        return np.array(_entropy(values, counts))
    return np.array(_asm(values, counts))


def _nan_entropy(x):
    """Compute the entropy of an array using a dask reduction operation."""
    return da.reduction(
        x,
        # mode chunk and combine funcs can be reused here
        chunk=_unique_with_counts_chunk,
        combine=_unique_with_counts_combine,
        aggregate=partial(_entropy_asm_agg, compute_entropy=True),
        dtype=F64,
        # Turn off concatenation to prevent dask from trying to concat the
        # dicts of variable length values and counts. Dask tries to concat
        # along the wrong axis, which causes errors.
        concatenate=False,
    )


def _nan_asm(x):
    """Compute the ASM of an array using a dask reduction operation.

    Angular second moment.
    """
    return da.reduction(
        x,
        # mode chunk and combine funcs can be reused here
        chunk=_unique_with_counts_chunk,
        combine=_unique_with_counts_combine,
        aggregate=partial(_entropy_asm_agg, compute_entropy=False),
        dtype=F64,
        # Turn off concatenation to prevent dask from trying to concat the
        # dicts of variable length values and counts. Dask tries to concat
        # along the wrong axis, which causes errors.
        concatenate=False,
    )


_ZONAL_STAT_FUNCS = {
    "asm": _nan_asm,
    "count": _nan_count,
    "entropy": _nan_entropy,
    "max": dask_nanmax,
    "mean": da.nanmean,
    "median": _nan_median,
    "min": dask_nanmin,
    "mode": _nan_mode,
    "std": da.nanstd,
    "sum": da.nansum,
    "unique": _nan_unique,
    "var": da.nanvar,
}
# The set of valid zonal function names/keys
ZONAL_STAT_FUNCS = frozenset(_ZONAL_STAT_FUNCS)


def _build_zonal_stats_data(data_raster, feat_raster, feat_labels, stats):
    nbands = data_raster.shape[0]
    feat_data = feat_raster.data
    # data will end up looking like:
    # {
    #   # band number
    #   1: {
    #     # Stat results
    #     "mean": [X, X, X], <- dask array
    #     "std": [X, X, X],
    #     ...
    #   },
    #   2: {
    #     # Stat results
    #     "mean": [X, X, X],
    #     "std": [X, X, X],
    #     ...
    #   },
    #   ...
    data = {}
    raster_data = get_raster(data_raster, null_to_nan=True).data
    for ibnd in range(nbands):
        ibnd += 1
        data[ibnd] = {}
        # Use range to keep band dimension intact
        band_data = raster_data[ibnd - 1 : ibnd]
        for f in stats:
            result_delayed = dask.delayed(ndmeasure.labeled_comprehension)(
                band_data,
                feat_data,
                feat_labels,
                _ZONAL_STAT_FUNCS[f],
                F64,
                np.nan,
            )
            data[ibnd][f] = da.from_delayed(
                result_delayed,
                feat_labels.shape,
                dtype=F64,
                meta=np.array([], dtype=F64),
            )
    return data


def _create_dask_range_index(start, stop):
    # dask.dataframe only allows dask.dataframe.index objects but doesn't have
    # a way to create them. this is a hack to create one using from_pandas.
    dummy = pd.DataFrame(
        {"tmp": np.zeros(stop - start, dtype="u1")},
        index=pd.RangeIndex(start, stop),
    )
    return dd.from_pandas(dummy, 1).index


def _build_zonal_stats_dataframe(zonal_data, nparts=None):
    bands = list(zonal_data)
    snames = list(zonal_data[bands[0]])
    n = zonal_data[bands[0]][snames[0]].size
    if nparts is None:
        # Get the number of partitions that dask thinks is reasonable. The data
        # arrays have chunks of size 1 so we need to rechunk later and then
        # repartition everything else in the dataframe to match.
        nparts = zonal_data[bands[0]][snames[0]].rechunk().npartitions

    df = None
    for bnd in bands:
        df_part = None
        band_data = zonal_data[bnd]
        band = da.full(n, bnd, dtype=I64)
        # We need to create an index because the concat operation later will
        # blindly paste in each dataframe's index. If an explicit index is not
        # set, the default is a range index from 0 to n. Thus the final
        # resulting dataframe would have identical indexes chained end-to-end:
        # [0, 1, ..., n-1, 0, 1, ..., n-1, 0, 1..., n-1]. By setting an index
        # we get [0, 1, ..., n, n+1, ..., n + n, ...].
        ind_start = n * (bnd - 1)
        ind_end = ind_start + n
        index = _create_dask_range_index(ind_start, ind_end)
        df_part = band.to_dask_dataframe("band", index=index).to_frame()
        # Repartition to match the data
        df_part = df_part.repartition(npartitions=nparts)
        index = index.repartition(npartitions=nparts)
        for name in snames:
            df_part[name] = (
                band_data[name].rechunk().to_dask_dataframe(name, index=index)
            )
        if df is None:
            df = df_part
        else:
            # Use interleave_partitions to keep partition and division info
            df = dd.concat([df, df_part], interleave_partitions=True)
    return df


def zonal_stats(features, data_raster, stats, raster_feature_values=None):
    """Apply stat functions to a raster based on a set of features.

    Parameters
    ----------
    features : str, Vector, Raster
        A `Vector` or path string pointing to a vector file or a categorical
        Raster. The vector features are used like cookie cutters to pull data
        from the `data_raster` bands. If `features` is a Raster, it must be an
        int dtype and have only one band.
    data_raster : Raster, str
        A `Raster` or path string pointing to a raster file. The data raster
        to pull data from and apply the stat functions to.
    stats : str, list of str
        A single string or list of strings corresponding to stat funcstions.
        These functions will be applied to the raster data for each of the
        features in `features`. Valid string values:

        'asm'
            Angular second moment. Applies -sum(P(g)**2) where P(g) gives the
            probability of g within the neighborhood.
        'count'
            Count valid cells.
        'entropy'
            Calculates the entropy. Applies -sum(P(g) * log(P(g))). See 'asm'
            above.
        'max'
            Find the maximum value.
        'mean'
            Calculate the mean.
        'median'
            Calculate the median value.
        'min'
            Find the minimum value.
        'mode'
            Compute the statistical mode of the data. In the case of a tie, the
            lowest value is returned.
        'std'
            Calculate the standard deviation.
        'sum'
            Calculate the sum.
        'unique'
            Count unique values.
        'var'
            Calculate the variance.
    raster_feature_values : sequence of ints, optional
        Unique values to be used when the `features` argument is a Raster. If
        `features` is a Raster and this is not provided the unique values in
        the raster will be calculated.

    Returns
    -------
    dask.dataframe.DataFrame
        A delayed dask DataFrame. The columns are the values in `stats` plus a
        column indicating the band the calculation was carried out on. Each row
        is the set of statistical calculations carried out on data pulled from
        `data_raster` based on the corresponding feature in `features`. NaN
        values indicate where a feature was outside of the raster or all data
        under the feature was null.

    """
    if is_str(features) or isinstance(features, Vector):
        features = get_vector(features)
    elif isinstance(features, Raster):
        if not is_int(features.dtype):
            raise TypeError("Feature raster must be an integer type.")
        if features.shape[0] > 1:
            raise ValueError("Feature raster must have only 1 band.")
    else:
        raise TypeError(
            "Could not understand features arg. Must be Vector, str or Raster"
        )
    data_raster = get_raster(data_raster)
    if is_str(stats):
        stats = [stats]
    elif isinstance(stats, Sequence):
        stats = list(stats)
        if not stats:
            raise ValueError("No stat functions provide")
    else:
        raise ValueError(f"Could not understand stats arg: {repr(stats)}")
    for stat in stats:
        if stat not in ZONAL_STAT_FUNCS:
            raise ValueError(f"Invalid stats function: {repr(stat)}")
    if isinstance(features, Raster):
        if features.crs != data_raster.crs:
            raise ValueError("Feature raster CRS must match data raster")
        if features.shape != data_raster.shape:
            raise ValueError("Feature raster shape must match data raster")

    feature_labels = None
    features_raster = None
    if isinstance(features, Vector):
        feature_labels = np.arange(1, len(features) + 1)
        features_raster = features.to_raster(data_raster)
    else:
        if raster_feature_values is None:
            (raster_feature_values,) = dask.compute(np.unique(features.data))
        else:
            raster_feature_values = np.atleast_1d(raster_feature_values)
            raster_feature_values = raster_feature_values[
                raster_feature_values > 0
            ]
        feature_labels = raster_feature_values
        features_raster = features

    data = _build_zonal_stats_data(
        data_raster, features_raster, feature_labels, stats
    )
    df = _build_zonal_stats_dataframe(data)
    return df


def _xy_to_rowcol_wrapper(x, y, affine):
    return np.stack(xy_to_rowcol(x, y, affine), axis=0)


def _extract_points(data, r, c, valid_mask):
    r, c, valid_mask = dask.compute(r, c, valid_mask)
    extracted = np.full((len(valid_mask),), np.nan, dtype=F64)
    extracted[valid_mask] = data[r[valid_mask], c[valid_mask]]
    return extracted


def _build_zonal_stats_data_from_points(data, mask, x, y, affine):
    r, c = da.blockwise(
        _xy_to_rowcol_wrapper,
        "zi",
        x,
        "i",
        y,
        "i",
        affine=affine,
        new_axes={"z": 2},
        dtype=np.int64,
    )
    _, rn, cn = data.shape
    r, c = dask.compute(r, c)
    valid_mask = (r >= 0) & (r < rn) & (c >= 0) & (c < cn)
    out = {
        i + 1: {"extracted": da.full(len(valid_mask), np.nan, dtype=F64)}
        for i in range(data.shape[0])
    }
    for i in range(data.shape[0]):
        extracted = da.full(len(valid_mask), np.nan, dtype=F64)
        extracted[valid_mask] = data.vindex[i, r[valid_mask], c[valid_mask]]
        # Mask out missing points within the valid zones
        exmask = da.zeros(len(valid_mask), dtype=bool)
        exmask[valid_mask] = mask.vindex[i, r[valid_mask], c[valid_mask]]
        extracted[exmask] = np.nan
        out[i + 1]["extracted"] = extracted
    return out


def extract_points_eager(
    points, raster, column_name="extracted", skip_validation=True
):
    """Extract the raster cell values using point features

    Note
    ----
    This function is partially eager. The x and y values for the target points
    are computed. The result is still a lazy dask DataFrame.

    This finds the grid cells that the points fall into and extracts the value
    at each point. The input feature will be partially computed to make sure
    that all of the geometries are points, unless `skip_validation` is set to
    `True`.

    Parameters
    ----------
    points : str, Vector
        The points to use for extracting data.
    raster : str, Raster
        The raster to pull data from.
    column_name : str, optional
        The column name to use for the extracted data points. Default is
        `"extracted"`.
    skip_validation : bool, optional
        If `True`, the input `points` is not validated to make sure that all
        features are points. This prevents partially computing the data.
        Default is `True`.

    Returns
    -------
    dask.dataframe.DataFrame
        The columns are "band" and `column_name. These are the band data was
        pulled from and the extracted value, respectively. NaN values in the
        extracted column are where there was missing data in the raster or
        the point was outside the raster's domain.

    """
    points = get_vector(points)
    raster = get_raster(raster)

    if not len(column_name):
        raise ValueError("column_name must not be empty")
    if (
        not skip_validation
        and not (points.geometry.geom_type == "Point").all().compute()
    ):
        raise TypeError("All geometries must be points.")

    gdf = points.to_crs(raster.crs).data
    x = gdf.geometry.x.to_dask_array()
    y = gdf.geometry.y.to_dask_array()
    r, c = raster.index(*dask.compute(x, y))
    nb, nr, nc = raster.shape
    valid = (r >= 0) & (r < nr) & (c >= 0) & (c < nc)
    n = len(valid)
    dfs = []
    for i in range(nb):
        bnd = i + 1
        index = _create_dask_range_index(n * i, n * bnd)
        df = (
            da.full(n, bnd, dtype=np.min_scalar_type(nb + 1))
            .to_dask_dataframe(index=index)
            .to_frame("band")
        )
        extracted = da.full(n, np.nan, dtype=F64)
        extracted[valid] = raster.data.vindex[i, r[valid], c[valid]]
        # Mask out missing points within the valid zones
        exmask = da.zeros(n, dtype=bool)
        exmask[valid] = raster.mask.vindex[i, r[valid], c[valid]]
        extracted[exmask] = np.nan
        df[column_name] = extracted.to_dask_dataframe(index=index)
        assert df.known_divisions
        dfs.append(df)
    df = dd.concat(dfs)
    return df



import operator
import unittest

import affine
import dask
import dask.array as da
import numpy as np
import pytest
import rasterio as rio
import rioxarray as riox
import xarray as xr
from dask_geopandas import GeoDataFrame

from raster_tools import Raster, band_concat
from raster_tools.dtypes import (
    DTYPE_INPUT_TO_DTYPE,
    F16,
    F32,
    F64,
    I8,
    I16,
    I32,
    I64,
    U8,
    U16,
    U32,
    U64,
    is_bool,
    is_float,
    is_int,
    is_scalar,
)
from raster_tools.masking import (
    get_default_null_value,
    reconcile_nullvalue_with_dtype,
)
from raster_tools.raster import (
    RasterQuadrantsResult,
    rowcol_to_xy,
    xy_to_rowcol,
)
from raster_tools.utils import is_strictly_decreasing, is_strictly_increasing
from tests.utils import arange_nd, arange_raster, assert_valid_raster

TEST_ARRAY = np.array(
    [
        [1, 3, 4, 4, 3, 2],
        [7, 3, 2, 6, 4, 6],
        [5, 8, 7, 5, 6, 6],
        [1, 4, 5, -1, 5, 1],
        [4, 7, 5, -1, 2, 6],
        [1, 2, 2, 1, 3, 4],
    ]
)


@pytest.mark.parametrize(
    "data",
    [
        np.ones((4, 4)),
        np.ones((1, 4, 4)),
        np.ones((1, 1)),
        arange_nd((100, 100)),
        arange_nd((10, 10, 10)),
        arange_nd((2, 10)),
        *[arange_nd((10, 10), dtype=t) for t in [I8, I32, U16, F16, F32]],
        np.where(TEST_ARRAY == -1, np.nan, TEST_ARRAY),
    ],
)
def test_raster_from_array(data):
    raster = Raster(data)
    assert_valid_raster(raster)
    assert raster._ds.raster.data is not data
    if data.ndim == 2:
        assert raster.shape[0] == 1
        assert raster.shape[1:] == data.shape
    else:
        assert raster.shape == data.shape
    assert raster.dtype == data.dtype
    if not np.isnan(data).any():
        assert raster.null_value is None
        assert not raster.mask.any()
        assert np.allclose(raster, data)
    else:
        mask = np.isnan(data)
        data_cp = data.copy()
        data_cp[mask] = get_default_null_value(data.dtype)
        assert raster.null_value == get_default_null_value(data.dtype)
        assert np.allclose(raster, data_cp)

    # Band dim starts at 1
    band = [1] if data.ndim == 2 else np.arange(data.shape[0]) + 1
    assert np.allclose(raster._ds.raster.band.data, band)
    # 2D coords are offset by 0.5
    assert np.allclose(
        raster._ds.raster.x.data, np.arange(data.shape[-1]) + 0.5
    )
    assert np.allclose(
        raster._ds.raster.y.data, (np.arange(data.shape[-2]) + 0.5)[::-1]
    )


@pytest.mark.parametrize(
    "data",
    [
        da.ones((4, 4)),
        da.ones((1, 4, 4)),
        da.ones((1, 1)),
        arange_nd((100, 100), mod=da),
        arange_nd((10, 10, 10), mod=da),
        arange_nd((2, 10), mod=da),
        *[
            arange_nd((10, 10), dtype=t, mod=da)
            for t in [I8, I32, U16, F16, F32]
        ],
        da.from_array(np.where(TEST_ARRAY == -1, np.nan, TEST_ARRAY)),
    ],
)
def test_raster_from_dask_array(data):
    assert dask.is_dask_collection(data)

    raster = Raster(data)
    assert_valid_raster(raster)
    assert raster._ds.raster.data is not data
    if data.ndim == 2:
        assert raster.shape == (1, *data.shape)
    else:
        assert raster.shape == data.shape
    assert raster.dtype == data.dtype
    assert dask.is_dask_collection(raster._ds.raster)
    assert dask.is_dask_collection(raster._ds.mask)
    if not is_float(data.dtype):
        assert raster.null_value is None
        assert not raster.mask.any().compute()
        assert np.allclose(raster, data.compute())
    else:
        mask = np.isnan(data)
        assert np.allclose(raster._ds.mask.data.compute(), mask.compute())
        data_cp = data.copy()
        data_cp[mask] = get_default_null_value(data.dtype)
        assert raster.null_value == get_default_null_value(data.dtype)
        assert np.allclose(raster, data_cp.compute())
    assert raster._ds.raster.dims == ("band", "y", "x")
    assert raster._ds.mask.dims == ("band", "y", "x")
    # Band dim starts at 1
    band = [1] if data.ndim == 2 else np.arange(data.shape[0]) + 1
    assert np.allclose(raster._ds.raster.band.data, band)
    # 2D coords are offset by 0.5
    assert np.allclose(
        raster._ds.raster.x.data, np.arange(data.shape[-1]) + 0.5
    )
    # y dim goes from high to low
    if data.shape[-2] > 1:
        assert is_strictly_decreasing(raster._ds.y)
    assert np.allclose(
        raster._ds.raster.y.data, (np.arange(data.shape[-2]) + 0.5)[::-1]
    )


@pytest.mark.parametrize(
    "xdata",
    [
        xr.DataArray(
            np.arange(20).reshape((4, 5)),
            coords=(np.arange(4)[::-1], np.arange(5)),
        ),
        xr.DataArray(
            np.arange(20.0).reshape((4, 5)),
            coords=(np.arange(4)[::-1], np.arange(5)),
        )
        .rio.write_nodata(np.nan)
        .rio.write_crs(3857),
        xr.DataArray(
            np.arange(20).reshape((4, 5)),
            # increasing y
            coords=(np.arange(4), np.arange(5)),
        ),
        xr.DataArray(
            np.arange(20).reshape((4, 5)),
            dims=("y", "x"),
            coords=(np.arange(4)[::-1], np.arange(5)),
        ),
        xr.DataArray(
            np.arange(40).reshape((2, 4, 5)),
            coords=([1, 2], np.arange(4)[::-1], np.arange(5)),
        ),
        xr.DataArray(
            np.arange(40).reshape((2, 4, 5)),
            dims=("band", "y", "x"),
            coords=([1, 2], np.arange(4)[::-1], np.arange(5)),
        ),
        xr.DataArray(
            np.arange(40).reshape((2, 4, 5)),
            dims=("band", "y", "x"),
            coords=([1, 2], np.arange(4)[::-1], np.arange(5)),
        ).rio.write_crs("EPSG:3857"),
        xr.DataArray(
            np.arange(20).reshape((1, 4, 5)),
            dims=("other", "lat", "lon"),
            coords=([1], np.arange(4)[::-1], np.arange(5)),
        ),
        # Dask
        xr.DataArray(
            da.from_array(np.arange(20).reshape((4, 5))),
            coords=(np.arange(4)[::-1], np.arange(5)),
        ),
        xr.DataArray(
            da.from_array(np.arange(20).reshape((1, 4, 5))),
            dims=("other", "lat", "lon"),
            coords=([1], np.arange(4)[::-1], np.arange(5)),
        ).chunk((1, 1, 1)),
    ],
)
def test_raster_from_dataarray(xdata):
    raster = Raster(xdata)
    assert_valid_raster(raster)
    assert raster._ds.raster is not xdata

    yc = xdata.dims[1] if xdata.ndim == 3 else xdata.dims[0]
    xc = xdata.dims[2] if xdata.ndim == 3 else xdata.dims[1]
    if is_strictly_increasing(xdata[yc]):
        xdata = xdata.isel({yc: slice(None, None, -1)})
    xdata = xdata.compute()
    assert raster.crs == xdata.rio.crs

    if xdata.ndim == 2:
        assert raster.shape[0] == 1
        assert raster.shape[1:] == xdata.shape
    else:
        assert raster.shape == xdata.shape
    assert raster.dtype == xdata.dtype
    if not is_float(xdata) or not np.isnan(xdata.data).any():
        if xdata.rio.nodata is None:
            assert raster.null_value is None
            assert not raster.mask.any().compute()
        elif np.isnan(xdata.rio.nodata):
            assert raster.null_value == get_default_null_value(xdata.dtype)
            assert np.allclose(raster.xmask, xdata == xdata.rio.nodata)
        else:
            assert raster.null_value == xdata.rio.nodata
            assert np.allclose(raster.xmask, xdata == xdata.rio.nodata)
        assert np.allclose(raster, xdata)
    elif np.isnan(xdata.data).any():
        mask = np.isnan(xdata.data)
        data_cp = xdata.data.copy()
        data_cp[mask] = get_default_null_value(xdata.dtype)
        assert raster.null_value == get_default_null_value(xdata.dtype)
        assert np.allclose(raster, data_cp)
    else:
        assert False
    # Band dim starts at 1
    band = [1] if xdata.ndim == 2 else np.arange(xdata.shape[0]) + 1
    assert np.allclose(raster._ds.band, band)
    assert np.allclose(raster._ds.x, xdata[xc])
    assert np.allclose(raster._ds.y, xdata[yc])


@pytest.mark.parametrize(
    "ds",
    [
        # Vanilla
        xr.Dataset(
            {
                "raster": (("y", "x"), arange_nd((10, 10))),
                "mask": (("y", "x"), np.zeros((10, 10), dtype=bool)),
            },
            coords={"y": np.arange(10)[::-1], "x": np.arange(10)},
        ),
        xr.Dataset(
            {
                "raster": (
                    ("y", "x"),
                    np.where(TEST_ARRAY == -1, np.nan, TEST_ARRAY),
                ),
                "mask": (("y", "x"), TEST_ARRAY == -1),
            },
            coords={
                "y": np.arange(TEST_ARRAY.shape[0])[::-1],
                "x": np.arange(TEST_ARRAY.shape[1]),
            },
        ),
        xr.Dataset(
            {
                "raster": (("band", "y", "x"), arange_nd((1, 10, 10))),
                "mask": (
                    ("y", "x"),
                    np.zeros((10, 10), dtype=bool),
                ),
            },
            coords={
                "band": [1],
                "y": np.arange(10)[::-1],
                "x": np.arange(10),
            },
            attrs={"_FillValue": 10},
        ),
        xr.Dataset(
            {
                "raster": (("y", "x"), arange_nd((10, 10))),
                "mask": (
                    ("band", "y", "x"),
                    np.zeros((1, 10, 10), dtype=bool),
                ),
            },
            coords={
                "band": [1],
                "y": np.arange(10)[::-1],
                "x": np.arange(10),
            },
        ),
        xr.Dataset(
            {
                "raster": (("lat", "lon"), arange_nd((10, 10))),
                "mask": (
                    ("lat", "lon"),
                    np.zeros((10, 10), dtype=bool),
                ),
            },
            coords={
                "lat": np.arange(10)[::-1],
                "lon": np.arange(10),
            },
        ),
        xr.Dataset(
            {
                "raster": (("band", "lat", "lon"), arange_nd((2, 10, 10))),
                "mask": (
                    ("band", "lat", "lon"),
                    np.zeros((2, 10, 10), dtype=bool),
                ),
            },
            coords={
                "band": [1, 2],
                "lat": np.arange(10)[::-1],
                "lon": np.arange(10),
            },
        ),
        # Lazy
        xr.Dataset(
            {
                "raster": (("y", "x"), arange_nd((10, 10))),
                "mask": (("y", "x"), np.zeros((10, 10), dtype=bool)),
            },
            coords={"y": np.arange(10)[::-1], "x": np.arange(10)},
        ).chunk(),
        xr.Dataset(
            {
                "raster": (("y", "x"), arange_nd((10, 10), mod=da)),
                "mask": (("y", "x"), np.zeros((10, 10), dtype=bool)),
            },
            coords={"y": np.arange(10)[::-1], "x": np.arange(10)},
        ),
        xr.Dataset(
            {
                "raster": (("y", "x"), arange_nd((10, 10))),
                "mask": (("y", "x"), da.zeros((10, 10), dtype=bool)),
            },
            coords={"y": np.arange(10)[::-1], "x": np.arange(10)},
        ),
        # flipped y dim
        xr.Dataset(
            {
                "raster": (("band", "y", "x"), arange_nd((1, 10, 10))),
                "mask": (
                    ("band", "y", "x"),
                    np.zeros((1, 10, 10), dtype=bool),
                ),
            },
            coords={
                "band": [1],
                "y": np.arange(10),
                "x": np.arange(10),
            },
        ),
        # CRS
        xr.Dataset(
            {
                "raster": (("band", "lat", "lon"), arange_nd((2, 10, 10))),
                "mask": (
                    ("band", "lat", "lon"),
                    np.zeros((2, 10, 10), dtype=bool),
                ),
            },
            coords={
                "band": [1, 2],
                "lat": np.arange(10)[::-1],
                "lon": np.arange(10),
            },
        ).rio.write_crs("EPSG:3857"),
    ],
)
def test_raster_from_dataset(ds):
    raster = Raster(ds)
    assert raster._ds is not ds
    assert_valid_raster(raster)
    assert raster.dtype == ds.raster.dtype
    assert raster.null_value == get_default_null_value(ds.raster.dtype)
    assert raster.crs == ds.rio.crs
    ds = ds.compute()
    yc = "y" if "y" in ds.dims else "lat"
    xc = "x" if "x" in ds.dims else "lon"
    if is_strictly_increasing(ds[yc]):
        ds = ds.isel({yc: slice(None, None, -1)})

    if ds.mask.any():
        assert np.allclose(
            raster,
            xr.where(
                ds.mask, get_default_null_value(ds.raster.dtype), ds.raster
            ),
        )
    else:
        assert np.allclose(raster._ds.raster, ds.raster)
    assert np.allclose(raster._ds.mask, ds.mask)
    assert np.allclose(raster._ds.x, ds[xc])
    assert np.allclose(raster._ds.y, ds[yc])


@pytest.mark.parametrize(
    "path",
    [
        "tests/data/raster/elevation.tif",
        "tests/data/raster/elevation_clipped.tif",
    ],
)
def test_raster_from_file(path):
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert_valid_raster(raster)
    assert raster.dtype == xdata.dtype
    if xdata.rio.nodata is None:
        assert raster.null_value is None
    else:
        assert np.allclose(raster.null_value, xdata.rio.nodata)
    assert raster.shape == xdata.shape
    if xdata.rio.nodata is not None and np.isnan(xdata.rio.nodata):
        assert np.allclose(
            raster,
            xr.where(
                xdata.mask,
                get_default_null_value(xdata.dtype),
                xdata,
            ),
        )
    else:
        assert np.allclose(raster, xdata)


@pytest.mark.parametrize(
    "data,error_type",
    [
        (np.ones(4), ValueError),
        (np.array(9), ValueError),
        (np.ones(1), ValueError),
        (da.ones(4), ValueError),
        # dims without coords
        (xr.DataArray(np.ones((4, 4))), ValueError),
    ],
)
def test_raster_from_any_errors(data, error_type):
    with pytest.raises(error_type):
        Raster(data)


@pytest.mark.parametrize(
    "raster_input,masked",
    [
        ("tests/data/raster/elevation_small.tif", True),
        ("tests/data/raster/elevation_clipped_small.tif", True),
        (np.ones((1, 3, 3)), False),
        (np.array([np.nan, 3.2, 4, 5]).reshape((2, 2)), True),
        (
            riox.open_rasterio(
                "tests/data/raster/elevation_clipped_small.tif"
            ).rio.write_nodata(None),
            False,
        ),
    ],
)
def test_property__masked(raster_input, masked):
    raster = Raster(raster_input)
    assert hasattr(raster, "_masked")
    assert raster._masked == masked


def test_property_values():
    raster = Raster("tests/data/raster/elevation_clipped_small.tif")
    assert hasattr(raster, "values")
    assert isinstance(raster.values, np.ndarray)
    assert raster.shape == raster.values.shape
    assert np.allclose(raster.values, raster._ds.raster.data.compute())


def test_property_null_value():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "null_value")
    assert raster.null_value == xdata.rio.nodata


def test_property_dtype():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "dtype")
    assert isinstance(raster.dtype, np.dtype)
    assert raster.dtype == xdata.dtype


def test_property_shape():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "shape")
    assert isinstance(raster.shape, tuple)
    assert raster.shape == xdata.shape


def test_property_crs():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "crs")
    assert isinstance(raster.crs, rio.crs.CRS)
    assert raster.crs == xdata.rio.crs


def test_property_affine():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "affine")
    assert isinstance(raster.affine, affine.Affine)
    assert raster.affine == xdata.rio.transform()


def test_property_resolution():
    path = "tests/data/raster/elevation_clipped_small.tif"
    raster = Raster(path)
    xdata = riox.open_rasterio(path)
    assert hasattr(raster, "resolution")
    assert isinstance(raster.resolution, tuple)
    assert raster.resolution == xdata.rio.resolution()


def test_property_xdata():
    rs = Raster("tests/data/raster/elevation_small.tif")
    assert hasattr(rs, "xdata")
    assert isinstance(rs.xdata, xr.DataArray)
    assert rs.xdata.identical(rs._ds.raster)


def test_property_data():
    rs = Raster("tests/data/raster/elevation_small.tif")
    assert hasattr(rs, "data")
    assert isinstance(rs.data, da.Array)
    assert rs.data is rs._ds.raster.data


def test_property_bounds():
    rs = Raster("tests/data/raster/elevation_small.tif")
    rds = rio.open("tests/data/raster/elevation_small.tif")
    assert hasattr(rs, "bounds")
    assert isinstance(rs.bounds, tuple)
    assert len(rs.bounds) == 4
    assert rs.bounds == tuple(rds.bounds)

    rs = Raster(
        np.array(
            [
                [0, 0],
                [0, 0],
            ]
        )
    )
    assert rs.bounds == (0, 0, 2, 2)


def test_property_mask():
    rs = Raster(np.arange(100).reshape((10, 10)) % 4).set_null_value(0)
    assert rs._ds.mask.data.sum().compute() > 0

    assert hasattr(rs, "mask")
    assert isinstance(rs.mask, dask.array.Array)
    assert np.allclose(rs.mask.compute(), rs._ds.mask.data.compute())


def test_property_xmask():
    rs = Raster(np.arange(100).reshape((10, 10)) % 4).set_null_value(0)
    rs._ds = rs._ds.rio.write_crs("EPSG:3857")
    assert rs.mask.sum().compute() > 0

    assert hasattr(rs, "xmask")
    assert isinstance(rs.xmask, xr.DataArray)
    assert np.allclose(rs.xmask.data.compute(), rs._ds.mask.data.compute())
    assert rs.xmask.rio.crs == rs.crs
    assert np.allclose(rs.xmask.x.data, rs.xdata.x.data)
    assert np.allclose(rs.xmask.y.data, rs.xdata.y.data)


@pytest.mark.parametrize("name", ["band", "x", "y"])
def test_properties_coords(name):
    rs = Raster("tests/data/raster/elevation_small.tif")

    assert hasattr(rs, name)
    assert isinstance(getattr(rs, name), np.ndarray)
    assert np.allclose(getattr(rs, name), rs.xdata[name].data)


@pytest.mark.parametrize(
    "rs",
    [
        Raster(da.ones((4, 100, 100), chunks=(1, 5, 5))),
        Raster(da.ones((100, 100), chunks=(5, 5))),
        Raster(np.ones((100, 100))),
        Raster("tests/data/raster/elevation.tif"),
    ],
)
def test_get_chunked_coords(rs):
    xc, yc = rs.get_chunked_coords()

    assert isinstance(xc, da.Array)
    assert isinstance(yc, da.Array)
    assert xc.dtype == rs._ds.x.dtype
    assert yc.dtype == rs._ds.y.dtype
    assert xc.ndim == 3
    assert yc.ndim == 3
    assert xc.shape == (1, 1, rs._ds.x.size)
    assert yc.shape == (1, rs._ds.y.size, 1)
    assert xc.chunks[:2] == ((1,), (1,))
    assert xc.chunks[2] == rs.data.chunks[2]
    assert yc.chunks[::2] == ((1,), (1,))
    assert yc.chunks[1] == rs.data.chunks[1]
    assert np.allclose(xc.ravel().compute(), rs._ds.x.data)
    assert np.allclose(yc.ravel().compute(), rs._ds.y.data)


_BINARY_ARITHMETIC_OPS = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.pow,
    operator.truediv,
    operator.floordiv,
    operator.mod,
]
_BINARY_COMPARISON_OPS = [
    operator.lt,
    operator.le,
    operator.gt,
    operator.ge,
    operator.eq,
    operator.ne,
]


@pytest.mark.parametrize("op", _BINARY_ARITHMETIC_OPS + _BINARY_COMPARISON_OPS)
@pytest.mark.parametrize("operand", [-2.0, -1, 0, 2, 3.0, True])
@pytest.mark.parametrize(
    "rs_type", [F16, F32, F64, I16, I32, I64, I8, U16, U32, U64, U8]
)
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:overflow")
def test_binary_ops_arithmetic_against_scalar(op, operand, rs_type):
    x = arange_nd((4, 5, 5), dtype=rs_type)
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    x = rs.values

    result = op(operand, rs)
    truth = op(operand, x)
    truth = np.where(mask, get_default_null_value(truth.dtype), truth)
    assert_valid_raster(result)
    assert result._masked
    assert result.crs == rs.crs
    assert np.allclose(result, truth, equal_nan=True)
    assert result.dtype == truth.dtype
    assert np.allclose(result.mask.compute(), rs.mask.compute())
    assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values
    if is_bool(result.dtype):
        assert is_bool(result.null_value)
        assert result.null_value == get_default_null_value(bool)
    # Make sure raising ints to negative ints raises TypeError
    if (
        op == operator.pow
        and is_int(rs.dtype)
        and is_int(operand)
        and operand < 0
        # Numpy promotes to float64 in this case
        and rs.dtype != U64
    ):
        with pytest.raises(TypeError):
            op(rs, operand)
    else:
        truth = op(x, operand)
        truth = np.where(mask, get_default_null_value(truth.dtype), truth)
        result = op(rs, operand)
        assert_valid_raster(result)
        assert result._masked
        assert result.dtype == truth.dtype
        assert result.crs == rs.crs
        assert np.allclose(result, truth, equal_nan=True)
        assert np.allclose(result.mask.compute(), rs.mask.compute())
        assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values


unknown_chunk_array = dask.array.ones((5, 5))
unknown_chunk_array = unknown_chunk_array[unknown_chunk_array > 0]


@pytest.mark.parametrize("op", _BINARY_ARITHMETIC_OPS + _BINARY_COMPARISON_OPS)
@pytest.mark.parametrize(
    "other,error",
    [
        ([1], False),
        (np.array([1]), False),
        (np.array([[1]]), False),
        (np.ones((5, 5)), False),
        (np.ones((1, 5, 5)), False),
        (np.ones((4, 5, 5)), False),
        (dask.array.ones((5, 5)), False),
        (np.zeros(4), True),
        (np.array([1, 1]), True),
        (unknown_chunk_array, True),
    ],
)
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:overflow")
@pytest.mark.filterwarnings("ignore:elementwise comparison failed")
def test_binary_ops_arithmetic_against_array(op, other, error):
    x = arange_nd((4, 5, 5), dtype=float)
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    data = rs.values

    if error:
        with pytest.raises(ValueError):
            op(rs, other)
        with pytest.raises(ValueError):
            op(other, rs)
    else:
        truth1 = op(data, other)
        truth2 = op(other, data)
        # Compute because dask throws an error when indexing with a mask
        truth1, truth2 = dask.compute(truth1, truth2)
        truth1[mask] = get_default_null_value(truth1.dtype)
        truth2[mask] = get_default_null_value(truth2.dtype)
        result = op(rs, other)
        assert_valid_raster(result)
        assert result.crs == rs.crs
        (truth1,) = dask.compute(truth1)
        assert np.allclose(result, truth1, equal_nan=True)
        assert np.allclose(result._ds.mask.compute(), rs._ds.mask.compute())
        assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values
        result = op(other, rs)
        assert_valid_raster(result)
        (truth2,) = dask.compute(truth2)
        assert np.allclose(result, truth2, equal_nan=True)
        assert np.allclose(result._ds.mask.compute(), rs._ds.mask.compute())
        assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values


@pytest.mark.parametrize("op", _BINARY_ARITHMETIC_OPS + _BINARY_COMPARISON_OPS)
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:overflow")
def test_binary_ops_arithmetic_against_raster(op):
    x = arange_nd((4, 5, 5), dtype=float)
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    data = rs.values
    data2 = np.ones_like(data, dtype=int) * 2
    rs2 = Raster(data2).set_crs("EPSG:3857")

    truth1 = op(data, data2)
    truth2 = op(data2, data)
    truth1[mask] = get_default_null_value(truth1.dtype)
    truth2[mask] = get_default_null_value(truth2.dtype)
    result = op(rs, rs2)
    assert_valid_raster(result)
    assert result.crs == rs.crs
    assert result._masked
    assert result.dtype == truth1.dtype
    assert np.allclose(result, truth1, equal_nan=True)
    assert np.allclose(result._ds.mask, rs._ds.mask)
    assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values
    result = op(rs2, rs)
    assert_valid_raster(result)
    assert result.crs == rs.crs
    assert result._masked
    assert result.dtype == truth2.dtype
    assert np.allclose(result, truth2, equal_nan=True)
    assert np.allclose(result._ds.mask, rs._ds.mask)
    assert np.all(result.xdata.spatial_ref == rs.xdata.spatial_ref).values


def test_binary_ops_arithmetic_inplace():
    data = np.arange(4 * 5 * 5).reshape((4, 5, 5))
    rs = Raster(data).set_crs("EPSG:3857")

    r = rs.copy()
    rr = r
    r += 3
    t = data + 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r -= 3
    t = data - 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r *= 3
    t = data * 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r **= 3
    t = data**3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r /= 3
    t = data / 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r //= 3
    t = data // 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs

    r = rs.copy()
    rr = r
    r %= 3
    t = data % 3
    assert np.allclose(r, t)
    assert rr is r
    assert rr.crs == rs.crs


_NP_UFUNCS = list(
    filter(
        lambda x: isinstance(x, np.ufunc)
        # not valid for rasters
        and x not in (np.isnat, np.matmul),
        map(lambda x: getattr(np, x), dir(np)),
    )
)
_NP_UFUNCS_NIN_SINGLE = list(filter(lambda x: x.nin == 1, _NP_UFUNCS))
_NP_UFUNCS_NIN_MULT = list(filter(lambda x: x.nin > 1, _NP_UFUNCS))
_UNSUPPORED_UFUNCS = frozenset((np.isnat, np.matmul))


@pytest.mark.parametrize("ufunc", list(_UNSUPPORED_UFUNCS))
def test_ufuncs_unsupported(ufunc):
    rs = Raster(np.arange(4 * 5 * 5).reshape((4, 5, 5)))
    with pytest.raises(TypeError):
        args = [rs for i in range(ufunc.nin)]
        ufunc(*args)


@pytest.mark.parametrize("ufunc", _NP_UFUNCS_NIN_SINGLE)
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
def test_ufuncs_single_input(ufunc):
    x = arange_nd((4, 5, 5), dtype=float)
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    assert rs.crs == "EPSG:3857"
    data = rs.values

    if ufunc == np.invert and is_float(data.dtype):
        with pytest.raises(TypeError):
            ufunc(rs)
        return
    truth = ufunc(data)
    result = ufunc(rs)
    if ufunc.nout == 1:
        assert_valid_raster(result)
        truth[mask] = get_default_null_value(truth.dtype)
        assert result._masked
        assert result.dtype == truth.dtype
        assert result.crs == rs.crs
        assert np.allclose(result, truth, equal_nan=True)
        assert np.allclose(result._ds.mask, rs._ds.mask, equal_nan=True)
    else:
        for r, t in zip(result, truth):
            assert_valid_raster(r)
            t[mask] = get_default_null_value(t.dtype)
            assert r._masked
            assert r.dtype == t.dtype
            assert r.crs == rs.crs
            assert np.allclose(r, t, equal_nan=True)
            assert np.allclose(r._ds.mask, rs._ds.mask, equal_nan=True)


@pytest.mark.parametrize("ufunc", _NP_UFUNCS_NIN_MULT)
@pytest.mark.parametrize("dtype", [int, float])
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:overflow")
def test_ufuncs_multiple_input_against_scalar(ufunc, dtype):
    x = arange_nd((4, 5, 5), dtype=dtype)
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    assert rs.crs == "EPSG:3857"
    extra = ufunc.nin - 1

    args = [rs] + [2 for i in range(extra)]
    args_np = [getattr(a, "values", a) for a in args]
    ufname = ufunc.__name__
    if (
        ufname.startswith("bitwise")
        or ufname.endswith("shift")
        or ufname in ("gcd", "lcm")
    ) and any(is_float(getattr(a, "dtype", a)) for a in args):
        with pytest.raises(TypeError):
            ufunc(*args)
    else:
        truth = ufunc(*args_np)
        result = ufunc(*args)
        if ufunc.nout == 1:
            assert_valid_raster(result)
            truth[mask] = get_default_null_value(truth.dtype)
            assert result._masked
            assert result.dtype == truth.dtype
            assert result.crs == rs.crs
            assert np.allclose(result, truth, equal_nan=True)
            assert np.allclose(result._ds.mask, rs._ds.mask, equal_nan=True)
        else:
            for r, t in zip(result, truth):
                assert_valid_raster(r)
                t[mask] = get_default_null_value(t.dtype)
                assert r._masked
                assert r.dtype == t.dtype
                assert r.crs == rs.crs
                assert np.allclose(r, t, equal_nan=True)
                assert np.allclose(r._ds.mask, rs._ds.mask, equal_nan=True)
    # Reflected
    args = args[::-1]
    args_np = args_np[::-1]
    if (
        ufname.startswith("bitwise")
        or ufname.endswith("shift")
        or ufname in ("gcd", "lcm")
    ) and any(is_float(getattr(a, "dtype", a)) for a in args):
        with pytest.raises(TypeError):
            ufunc(*args)
    elif ufname == "ldexp" and is_float(args[-1].dtype):
        with pytest.raises(TypeError):
            ufunc(*args)
    else:
        truth = ufunc(*args_np)
        result = ufunc(*args)
        if ufunc.nout == 1:
            assert_valid_raster(result)
            truth[mask] = get_default_null_value(truth.dtype)
            assert result._masked
            assert result.dtype == truth.dtype
            assert result.crs == rs.crs
            assert np.allclose(result, truth, equal_nan=True)
            assert np.allclose(result._ds.mask, rs._ds.mask, equal_nan=True)
        else:
            for r, t in zip(result, truth):
                assert_valid_raster(r)
                t[mask] = get_default_null_value(t.dtype)
                assert r._masked
                assert r.dtype == t.dtype
                assert r.crs == rs.crs
                assert np.allclose(r, t, equal_nan=True)
                assert np.allclose(r._ds.mask, rs._ds.mask, equal_nan=True)


@pytest.mark.parametrize("ufunc", _NP_UFUNCS_NIN_MULT)
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:overflow")
def test_ufuncs_multiple_input_against_raster(ufunc):
    x = arange_nd((4, 5, 5))
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    assert rs.crs == "EPSG:3857"
    extra = ufunc.nin - 1
    other = Raster(np.ones_like(x) * 2)
    other.xdata.data[0, 0, 0] = 25
    other = other.set_null_value(25)
    assert other._ds.mask.data.compute().sum() == 1
    mask = rs._ds.mask | other._ds.mask

    args = [rs] + [other.copy() for i in range(extra)]
    args_np = [a.values for a in args]

    if ufunc.__name__.startswith("bitwise") and any(
        is_float(getattr(a, "dtype", type(a))) for a in args
    ):
        with pytest.raises(TypeError):
            ufunc(*args)
    else:
        truth = ufunc(*args_np)
        result = ufunc(*args)
        if ufunc.nout == 1:
            assert_valid_raster(result)
            truth[mask] = get_default_null_value(truth.dtype)
            assert result._masked
            assert result.dtype == truth.dtype
            assert result.crs == rs.crs
            assert np.allclose(result, truth, equal_nan=True)
            assert np.allclose(result._ds.mask, mask, equal_nan=True)
        else:
            for r, t in zip(result, truth):
                assert_valid_raster(r)
                t[mask] = get_default_null_value(t.dtype)
                assert r._masked
                assert r.dtype == t.dtype
                assert r.crs == rs.crs
                assert np.allclose(r, t, equal_nan=True)
                assert np.allclose(r._ds.mask, mask, equal_nan=True)
    # Test reflected
    args = args[::-1]
    args_np = args_np[::-1]
    if ufunc.__name__.startswith("bitwise") and any(
        is_float(getattr(a, "dtype", type(a))) for a in args
    ):
        with pytest.raises(TypeError):
            ufunc(*args)
    else:
        truth = ufunc(*args_np)
        result = ufunc(*args)
        if ufunc.nout == 1:
            assert_valid_raster(result)
            truth[mask] = get_default_null_value(truth.dtype)
            assert result.crs == rs.crs
            assert np.allclose(result, truth, equal_nan=True)
            assert np.allclose(result._ds.mask, mask)
        else:
            for r, t in zip(result, truth):
                assert_valid_raster(r)
                t[mask] = get_default_null_value(t.dtype)
                assert r.crs == rs.crs
                assert np.allclose(r, t, equal_nan=True)
                assert np.allclose(r._ds.mask, mask, equal_nan=True)


def test_invert():
    x = arange_nd((4, 5, 5))
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    data = rs._ds.raster.data.compute()
    mask = rs._ds.mask.data.compute()
    assert rs.null_value == 0
    assert rs._masked
    assert rs.crs == "EPSG:3857"

    result = ~rs
    truth = np.invert(data)
    truth[mask] = get_default_null_value(truth.dtype)
    assert_valid_raster(result)
    assert result._masked
    assert result.dtype == truth.dtype
    assert (result).crs == rs.crs
    assert np.allclose(np.invert(rs), truth)
    assert np.allclose(result, truth)
    truth = np.invert(data.astype(bool))
    truth[mask] = get_default_null_value(truth.dtype)
    assert np.allclose(np.invert(rs.astype(bool)), truth)
    with pytest.raises(TypeError):
        ~rs.astype(float)


@pytest.mark.parametrize(
    "func",
    [
        np.all,
        np.any,
        np.max,
        np.mean,
        np.min,
        np.prod,
        np.std,
        np.sum,
        np.var,
    ],
)
def test_reductions(func):
    data = np.arange(3 * 5 * 5).reshape((3, 5, 5)) - 25
    data = data.astype(float)
    rs = Raster(data).set_null_value(4)
    rs.data[:, :2, :2] = np.nan
    valid = ~(rs._ds.mask.data | np.isnan(rs._ds.raster.data)).compute()

    fname = func.__name__
    if fname in ("amin", "amax"):
        # Drop leading 'a'
        fname = fname[1:]
    assert hasattr(rs, fname)
    assert not np.allclose(valid, rs._ds.mask.data.compute())
    truth = func(data[valid])
    for result in (func(rs), getattr(rs, fname)()):
        assert isinstance(result, dask.array.Array)
        assert result.size == 1
        assert result.shape == ()
        if fname in ("all", "any"):
            assert isinstance(result.compute(), (bool, np.bool_))
        else:
            assert is_scalar(result.compute())
        assert np.allclose(result, truth, equal_nan=True)
    if fname in ("all", "any"):
        rs = rs > 0
        data = data > 0
        truth = func(data[valid])
        assert np.allclose(func(rs), truth, equal_nan=True)
        assert np.allclose(getattr(rs, fname)(), truth, equal_nan=True)


@pytest.mark.parametrize(
    "other",
    [
        5,
        [9],
        (7,),
        (1, 2, 3, 4),
        [1, 2, 3, 4],
        range(1, 5),
        np.array([1, 2, 3, 4]),
        da.asarray([1, 2, 3, 4]),
    ],
)
@pytest.mark.parametrize("ufunc", _NP_UFUNCS_NIN_MULT)
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_bandwise(ufunc, other):
    x = arange_nd((4, 5, 5))
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), 0, rs._ds.raster
    )
    rs = rs.set_null_value(0).set_crs("EPSG:3857")
    data = rs.values
    mask = rs.mask.compute()
    assert rs.null_value == 0
    assert rs._masked
    assert rs.crs == "EPSG:3857"
    assert hasattr(rs, "bandwise")
    tother = dask.compute(np.atleast_1d(other).reshape((-1, 1, 1)))[0]

    truth = ufunc(data, tother)
    result = ufunc(rs.bandwise, other)
    if ufunc.nout == 1:
        assert_valid_raster(result)
        truth[mask] = get_default_null_value(truth.dtype)
        assert result._masked
        assert truth.dtype == result.dtype
        assert result.crs == rs.crs
        assert np.allclose(result, truth, equal_nan=True)
        assert np.allclose(result._ds.mask, mask)
    else:
        for r, t in zip(result, truth):
            assert_valid_raster(r)
            t[mask] = get_default_null_value(t.dtype)
            assert r._masked
            assert t.dtype == r.dtype
            assert r.crs == rs.crs
            assert np.allclose(r, t, equal_nan=True)
            assert np.allclose(r._ds.mask, mask)
    # Reflected
    truth = ufunc(tother, data)
    result = ufunc(other, rs.bandwise)
    if ufunc.nout == 1:
        assert_valid_raster(result)
        truth[mask] = get_default_null_value(truth.dtype)
        assert result._masked
        assert truth.dtype == result.dtype
        assert result.crs == rs.crs
        assert np.allclose(result, truth, equal_nan=True)
        assert np.allclose(result._ds.mask, mask)
    else:
        for r, t in zip(result, truth):
            assert_valid_raster(r)
            t[mask] = get_default_null_value(t.dtype)
            assert r._masked
            assert t.dtype == r.dtype
            assert r.crs == rs.crs
            assert np.allclose(r, t, equal_nan=True)
            assert np.allclose(r._ds.mask, mask)


@pytest.mark.parametrize(
    "other",
    [
        [1, 1],
        (1, 1),
        range(0, 3),
        np.ones(3),
        np.ones((2, 2)),
        np.ones((5, 5)),
        unknown_chunk_array,
    ],
)
def test_bandwise_errors(other):
    rs = Raster(arange_nd((4, 5, 5)))

    with pytest.raises(ValueError):
        rs.bandwise * other


def test_round():
    data = np.arange(5 * 5).reshape((5, 5)).astype(float)
    data += np.linspace(0, 4, 5 * 5).reshape((5, 5))
    rs = Raster(data).set_crs(3857)

    assert hasattr(rs, "round")
    truth = np.round(data)
    assert_valid_raster(rs.round())
    assert isinstance(rs.round(), Raster)
    assert np.allclose(rs.round(), truth)
    truth = np.round(data, decimals=2)
    assert np.allclose(rs.round(2), truth)
    assert rs.round().crs == rs.crs

    rs = Raster(data).set_crs(3857).set_null_value(24.5)
    truth = np.round(data)
    truth[rs.xmask.values[0]] = 24.5
    assert_valid_raster(rs.round())
    assert isinstance(rs.round(), Raster)
    assert np.allclose(rs.round(), truth)
    truth = np.round(data, decimals=2)
    truth[rs.xmask.values[0]] = 24.5
    assert np.allclose(rs.round(2), truth)
    assert rs.round().crs == rs.crs


@pytest.mark.filterwarnings("ignore:The null value ")
@pytest.mark.parametrize("type_code,dtype", list(DTYPE_INPUT_TO_DTYPE.items()))
@pytest.mark.parametrize(
    "rs",
    [
        Raster("tests/data/raster/elevation_small.tif"),
        Raster(np.arange(100).reshape((10, 10))).set_null_value(99),
    ],
)
def test_astype(rs, type_code, dtype):
    result = rs.astype(type_code)
    assert_valid_raster(result)
    assert result.dtype == dtype
    assert result.eval().dtype == dtype
    assert result.crs == rs.crs
    assert result.null_value == reconcile_nullvalue_with_dtype(
        rs.null_value, dtype
    )

    result = rs.astype(dtype)
    assert_valid_raster(rs.astype(dtype))
    assert result.dtype == dtype
    assert result.eval().dtype == dtype
    assert result.crs == rs.crs
    assert result.null_value == reconcile_nullvalue_with_dtype(
        rs.null_value, dtype
    )


@pytest.mark.filterwarnings("ignore:The null value ")
def test_astype_wrong_type_codes():
    rs = Raster("tests/data/raster/elevation_small.tif")
    with pytest.raises(ValueError):
        rs.astype("not float32")
    with pytest.raises(ValueError):
        rs.astype("other")


def test_copy():
    rs = Raster("tests/data/raster/elevation_clipped_small.tif")
    copy = rs.copy()
    assert_valid_raster(copy)
    assert rs is not copy
    assert rs._ds is not copy._ds
    assert rs._ds.equals(copy._ds)
    assert np.allclose(rs, copy)
    # make sure a deep copy has occurred
    copy._ds.raster.data[0, -1, -1] = 0
    assert not np.allclose(rs, copy)


def test_set_crs():
    rs = Raster("tests/data/raster/elevation_small.tif")
    assert rs.crs != 4326

    rs4326 = rs.set_crs(4326)
    assert rs4326.crs != rs.crs
    assert rs4326.crs == 4326
    assert np.allclose(rs.values, rs4326.values)


@pytest.mark.filterwarnings("ignore:The null value")
def test_set_null_value():
    rs = Raster("tests/data/raster/elevation_clipped_small.tif")
    assert rs.null_value is not None
    truth = rs.values
    mask = rs._ds.mask.data.compute()
    ndv = rs.null_value
    rs2 = rs.set_null_value(0)
    assert_valid_raster(rs2)
    truth[mask] = 0
    assert rs.null_value == ndv
    assert rs.xdata.rio.nodata == ndv
    assert rs2.xdata.rio.nodata == 0
    assert np.allclose(rs2, truth)
    assert rs2.crs == rs.crs

    rs = Raster(np.arange(100).reshape((10, 10)))
    assert rs.null_value is None
    assert rs._ds.raster.rio.nodata is None
    rs2 = rs.set_null_value(99)
    assert rs2.null_value == 99
    assert rs2._ds.raster.rio.nodata == 99
    assert np.allclose(rs2._ds.mask.values, rs2._ds.raster.values == 99)
    assert rs2.crs == rs2.crs

    rs = Raster("tests/data/raster/elevation_small.tif")
    nv = rs.null_value
    rs2 = rs.set_null_value(None)
    assert rs.null_value == nv
    assert rs2.null_value is None
    assert not rs2._ds.mask.values.any()
    assert np.allclose(rs, rs2)
    assert rs2.crs == rs2.crs

    rs = Raster("tests/data/raster/elevation_clipped_small.tif").astype(int)
    assert rs.dtype == np.dtype(int)
    rs2 = rs.set_null_value(get_default_null_value(float))
    assert rs2.dtype == np.dtype(float)
    assert rs2.crs == rs2.crs


def test_replace_null():
    fill_value = 0
    rs = Raster("tests/data/raster/elevation_clipped_small.tif")
    rsnp = rs.values
    rsnp_replaced = rsnp.copy()
    mask = rsnp == rs.null_value
    rsnp_replaced[mask] = fill_value
    rs = rs.replace_null(fill_value)
    assert np.allclose(rs.values, rsnp_replaced)
    assert rs.null_value == fill_value
    assert np.allclose(rs._ds.mask.values, mask)

    with pytest.raises(TypeError):
        rs.replace_null(None)


@pytest.mark.filterwarnings("ignore:The null value")
def test_where():
    rs = Raster("tests/data/raster/elevation_small.tif")
    c = rs > 1100

    r = rs.where(c, 0)
    assert_valid_raster(r)
    rsnp = np.asarray(rs)
    truth = np.where(rsnp > 1100, rsnp, 0)
    assert np.allclose(r, truth)
    assert np.allclose(
        rs.where(c, "tests/data/raster/elevation_small.tif"), rs
    )
    assert r.crs == rs.crs

    c = c.astype(int)
    r = rs.where(c, 0)
    assert_valid_raster(r)
    assert np.allclose(r, truth)

    assert rs._masked
    assert r._masked
    assert rs.crs is not None
    assert r.crs is not None
    assert r.crs == rs.crs
    assert r.null_value == get_default_null_value(r.dtype)

    with pytest.raises(TypeError):
        cf = c.astype(float)
        rs.where(cf, 0)
    with pytest.raises(TypeError):
        rs.where(c, None)


def test_to_null_mask():
    rs = Raster("tests/data/raster/elevation_clipped_small.tif")
    nv = rs.null_value
    rsnp = rs.values
    truth = rsnp == nv
    assert rs.null_value is not None
    nmask = rs.to_null_mask()
    assert_valid_raster(nmask)
    assert nmask.null_value is None
    assert np.allclose(nmask, truth)
    # Test case where no null values
    rs = Raster("tests/data/raster/elevation_small.tif")
    truth = np.full(rs.shape, False, dtype=bool)
    nmask = rs.to_null_mask()
    assert_valid_raster(nmask)
    assert nmask.null_value is None
    assert np.allclose(nmask, truth)


class TestEval(unittest.TestCase):
    def test_eval(self):
        rs = Raster("tests/data/raster/elevation_small.tif")
        rsnp = rs.xdata.values
        rs += 2
        rsnp += 2
        rs -= rs
        rsnp -= rsnp
        rs *= -1
        rsnp *= -1
        result = rs.eval()
        assert_valid_raster(result)
        # Make sure new raster returned
        self.assertIsNot(rs, result)
        self.assertIsNot(rs._ds, result._ds)
        # Make sure that original raster is still lazy
        assert np.allclose(result, rsnp)


def test_get_bands():
    rs = Raster(np.arange(4 * 100 * 100).reshape((4, 100, 100)))
    rsnp = rs.values
    r = rs.get_bands(1)
    assert_valid_raster(r)
    assert np.allclose(r, rsnp[:1])
    r = rs.get_bands(2)
    assert_valid_raster(r)
    assert np.allclose(r, rsnp[1:2])
    r = rs.get_bands(3)
    assert_valid_raster(r)
    assert np.allclose(r, rsnp[2:3])
    r = rs.get_bands(4)
    assert_valid_raster(r)
    assert np.allclose(r, rsnp[3:4])
    for bands in [[1], [1, 2], [1, 1], [3, 1, 2], [4, 3, 2, 1]]:
        np_bands = [i - 1 for i in bands]
        result = rs.get_bands(bands)
        assert_valid_raster(result)
        assert np.allclose(result, rsnp[np_bands])
        bnd_dim = list(range(1, len(bands) + 1))
        assert np.allclose(result.xdata.band, bnd_dim)

    assert len(rs.get_bands(1).shape) == 3

    for bands in [0, 5, [1, 5], [0]]:
        with pytest.raises(IndexError):
            rs.get_bands(bands)
    with pytest.raises(ValueError):
        rs.get_bands([])
    with pytest.raises(TypeError):
        rs.get_bands(1.0)


def test_burn_mask():
    x = arange_nd((1, 5, 5))
    rs = Raster(x)
    rs._ds["raster"] = xr.where(
        (0 <= rs._ds.raster) & (rs._ds.raster < 10), -999, rs._ds.raster
    )
    rs = rs.set_null_value(-999).set_crs("EPSG:3857")
    data = rs.values
    assert rs.null_value == -999
    assert rs._masked
    assert rs.crs == "EPSG:3857"
    true_mask = data < 10
    true_state = data.copy()
    true_state[true_mask] = -999
    assert np.allclose(true_mask, rs._ds.mask)
    assert np.allclose(true_state, rs)

    rs._ds.raster.data = data
    assert np.allclose(rs, data)
    assert_valid_raster(rs.burn_mask())
    assert np.allclose(rs.burn_mask(), true_state)
    assert rs.burn_mask().crs == rs.crs

    data = arange_nd((1, 5, 5))
    rs = Raster(data)
    rs._ds["raster"] = xr.where(
        (20 <= rs._ds.raster) & (rs._ds.raster < 26), 999, rs._ds.raster
    )
    rs = rs.set_null_value(999).set_crs("EPSG:3857")
    rs = rs > 15
    assert rs.dtype == np.dtype(bool)
    nv = get_default_null_value(bool)
    assert rs.null_value == nv
    true_state = data > 15
    true_state = np.where(data >= 20, nv, true_state)
    assert np.allclose(rs.burn_mask(), true_state)


@pytest.mark.parametrize("index", [(0, 0), (0, 1), (1, 0), (-1, -1), (1, -1)])
@pytest.mark.parametrize("offset", ["center", "ul", "ll", "ur", "lr"])
def test_xy(index, offset):
    rs = Raster("tests/data/raster/elevation_small.tif")
    i, j = index
    if i == -1:
        i = rs.shape[1] - 1
    if j == -1:
        j = rs.shape[2] - 1

    T = rs.affine
    assert rs.xy(i, j, offset) == rio.transform.xy(T, i, j, offset=offset)


@pytest.mark.parametrize(
    "x,y",
    [
        (0, 0),
        (-47848.38283538996, 65198.938461863494),
        (-47278.38283538996, 65768.9384618635),
        (-44878.38283538996, 68168.9384618635),
        (-46048.38283538996, 66308.9384618635),
        (-46048.0, 66328.1),
        (-46025.0, 66328.1),
    ],
)
def test_index(x, y):
    rs = Raster("tests/data/raster/elevation_small.tif")
    T = rs.affine

    assert rs.index(x, y) == rio.transform.rowcol(T, x, y)


@pytest.mark.parametrize("offset_name", ["center", "ul", "ur", "ll", "lr"])
def test_rowcol_to_xy(offset_name):
    rs = Raster("tests/data/raster/elevation_small.tif")
    affine = rs.affine
    _, ny, nx = rs.shape
    R, C = np.mgrid[:ny, :nx]
    r = R.ravel()
    c = C.ravel()
    xy_rio = rio.transform.xy(affine, r, c, offset=offset_name)
    result = rowcol_to_xy(r, c, affine, offset_name)
    assert np.allclose(xy_rio, result)


def test_xy_to_rowcol():
    rs = Raster("tests/data/raster/elevation_small.tif")
    affine = rs.affine
    X, Y = np.meshgrid(rs.x, rs.y)
    x = X.ravel()
    y = Y.ravel()
    rc_rio = rio.transform.rowcol(affine, x, y)
    result = xy_to_rowcol(x, y, affine)
    assert np.allclose(rc_rio, result)
    assert result[0].dtype == np.dtype(int)
    assert result[1].dtype == np.dtype(int)


def _compare_raster_to_vectorized(rs, df):
    data = rs.values
    x = rs.xdata.x.values
    y = rs.xdata.y.values
    for dfrow in df.itertuples():
        value = dfrow.value
        band = dfrow.band
        row = dfrow.row
        col = dfrow.col
        p = dfrow.geometry
        assert band >= 1
        assert value == data[band - 1, row, col]
        assert x[col] == p.x
        assert y[row] == p.y


def test_to_vector():
    data = np.array(
        [
            [
                [0, 1, 1, 2],
                [0, 1, 2, 2],
                [0, 0, 1, 0],
                [0, 1, 3, 0],
            ],
            [
                [3, 2, 2, 1],
                [0, 1, 2, 1],
                [0, 0, 2, 0],
                [0, 2, 4, 5],
            ],
        ]
    )
    count = np.sum(data > 0)
    rs = Raster(data).set_null_value(0)
    ddf = rs.to_vector()
    df = ddf.compute()

    assert isinstance(ddf, GeoDataFrame)
    assert ddf.crs == rs.crs
    assert len(df) == count
    assert np.all(ddf.columns == ["value", "band", "row", "col", "geometry"])
    _compare_raster_to_vectorized(rs, df)

    rs = Raster("tests/data/raster/elevation_small.tif")
    rs = band_concat((rs, rs + 100))
    data = rs.values
    rs = rs.chunk((1, 20, 20))
    ddf = rs.to_vector()
    df = ddf.compute()

    assert rs.data.npartitions == 50
    assert rs._ds.mask.values.sum() == 0
    assert ddf.npartitions == rs.data.npartitions
    assert ddf.crs == rs.crs
    assert len(df) == rs.data.size
    assert np.all(ddf.columns == ["value", "band", "row", "col", "geometry"])
    _compare_raster_to_vectorized(rs, df)

    # make sure that empty (all-null) chunks are handled
    data = np.array(
        [
            [
                [0, 0, 1, 2],
                [0, 0, 2, 2],
                [0, 0, 1, 0],
                [0, 1, 3, 0],
            ]
        ]
    )
    count = np.sum(data > 0)
    rs = Raster(data).set_null_value(0).chunk((1, 2, 2))
    ddf = rs.to_vector()
    df = ddf.compute()

    assert len(df) == count
    assert df["value"].dtype == data.dtype
    assert df["band"].dtype == np.dtype(int)
    assert df["row"].dtype == np.dtype(int)
    assert df["col"].dtype == np.dtype(int)
    _compare_raster_to_vectorized(rs, df)


@pytest.mark.parametrize(
    "raster",
    [
        Raster("tests/data/raster/elevation.tif"),
        arange_raster((3, 35, 50))
        .set_crs("EPSG:3857")
        .remap_range([(0, 10, 0), (3490, 3500, 0), (5200, 5210, 0)])
        .set_null_value(0),
        arange_raster((3, 1, 1)),
    ],
)
def test_to_quadrants(raster):
    nb, ny, nx = raster.shape
    assert hasattr(raster, "to_quadrants")
    quads = raster.to_quadrants()
    assert isinstance(quads, RasterQuadrantsResult)
    assert len(quads) == 4
    assert hasattr(quads, "nw")
    assert hasattr(quads, "ne")
    assert hasattr(quads, "sw")
    assert hasattr(quads, "se")
    for q in quads:
        assert_valid_raster(q)
        assert q.crs == raster.crs
        assert q.null_value == raster.null_value
        assert len(q.band) == nb
        assert np.allclose(q.band, raster.band)
    assert quads.nw == quads[0]
    assert quads.ne == quads[1]
    assert quads.sw == quads[2]
    assert quads.se == quads[3]
    assert quads.nw.x.size + quads.ne.x.size == nx
    assert quads.sw.x.size + quads.se.x.size == nx
    assert quads.nw.y.size + quads.sw.y.size == ny
    assert quads.ne.y.size + quads.se.y.size == ny
    assert np.allclose(np.concatenate([quads.nw.x, quads.ne.x]), raster.x)
    assert np.allclose(np.concatenate([quads.sw.x, quads.se.x]), raster.x)
    assert np.allclose(np.concatenate([quads.nw.y, quads.sw.y]), raster.y)
    assert np.allclose(np.concatenate([quads.ne.y, quads.se.y]), raster.y)
    xreconstructed = xr.concat(
        [
            xr.concat([quads.nw.xdata, quads.ne.xdata], dim="x"),
            xr.concat([quads.sw.xdata, quads.se.xdata], dim="x"),
        ],
        dim="y",
    )
    assert np.allclose(xreconstructed, raster.xdata)
    xmask_reconstructed = xr.concat(
        [
            xr.concat([quads.nw.xmask, quads.ne.xmask], dim="x"),
            xr.concat([quads.sw.xmask, quads.se.xmask], dim="x"),
        ],
        dim="y",
    )
    assert np.allclose(xmask_reconstructed, raster.xmask)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from numba import generated_jit, jit, types

__all__ = [
    "nan_unique_count_jit",
    "nanargmax_jit",
    "nanargmin_jit",
    "nanasm_jit",
    "nanentropy_jit",
    "nanmax_jit",
    "nanmean_jit",
    "nanmedian_jit",
    "nanmin_jit",
    "nanmode_jit",
    "nanstd_jit",
    "nansum_jit",
    "nanvar_jit",
]


ngjit = jit(nopython=True, nogil=True)


# Numba has a long standing bug where scalars are not handled properly in
# several numpy functions. This is a workaround. It generates the proper
# implementation based on the input type. This func can be used to properly
# handle scalar inputs in funcs that would otherwise fail to compile for scalar
# inputs like np.nanmean.
# See: https://github.com/numba/numba/issues/4202
@generated_jit(nopython=True, nogil=True)
def _atleast_1d(x):
    if x in types.number_domain:
        return lambda x: np.array([x])
    return lambda x: np.atleast_1d(x)


@ngjit
def nanmin_jit(x):
    return np.nanmin(x)


@ngjit
def nanmax_jit(x):
    return np.nanmax(x)


@ngjit
def nanmean_jit(x):
    x = _atleast_1d(x)
    return np.nanmean(x)


@ngjit
def nanmedian_jit(x):
    x = _atleast_1d(x)
    return np.nanmedian(x)


@ngjit
def nansum_jit(x):
    x = _atleast_1d(x)
    return np.nansum(x)


@ngjit
def nanvar_jit(x):
    x = _atleast_1d(x)
    return np.nanvar(x)


@ngjit
def nanstd_jit(x):
    x = _atleast_1d(x)
    return np.nanstd(x)


@ngjit
def nan_unique_count_jit(x):
    x = _atleast_1d(x)
    # Create set of floats. {1.0} is a hack to tell numba.jit what type the set
    # contains
    s = {1.0}
    s.clear()
    for v in x.ravel():
        if not np.isnan(v):
            s.add(v)
    return len(s)


@ngjit
def nanmode_jit(x):
    # Returns NaN if all values are NaN.
    x = _atleast_1d(x)
    c = {}
    x = x.ravel()
    j = 0
    for i in range(x.size):
        v = x[i]
        if not np.isnan(v):
            if v in c:
                c[v] += 1
            else:
                c[v] = 1
            x[j] = v
            j += 1
    vals = x[:j]
    if len(vals) == 0:
        return np.nan
    vals.sort()
    cnts = np.empty(len(vals), dtype=types.uint64)
    for i in range(len(vals)):
        cnts[i] = c[vals[i]]
    return vals[np.argmax(cnts)]


@ngjit
def nanentropy_jit(x):
    x = _atleast_1d(x)
    c = {}
    n = 0
    for v in x.ravel():
        if not np.isnan(v):
            if v in c:
                c[v] += 1
            else:
                c[v] = 1
            n += 1
    entr = 0.0
    if n > 0:
        frac = 1.0 / n
        for cnt in c.values():
            p = cnt * frac
            entr -= p * np.log(p)
    return entr


@ngjit
def nanasm_jit(x):
    x = _atleast_1d(x)
    c = {}
    n = 0
    for v in x.ravel():
        if not np.isnan(v):
            if v in c:
                c[v] += 1
            else:
                c[v] = 1
            n += 1
    asm = 0.0
    if n > 0:
        frac = 1.0 / n
        for cnt in c.values():
            p = cnt * frac
            asm += p * p
    return asm


@ngjit
def nanargmin_jit(x):
    # Get argmin. If empty give -1, if all nan, give -2. This differs from
    # numpy which throws an error.
    x = _atleast_1d(x)
    if x.size == 0:
        return -1
    nan_mask = np.isnan(x)
    if nan_mask.all():
        return -2
    x = x.ravel()
    nan_mask = nan_mask.ravel()
    j = 0
    min_val = np.inf
    for i in range(x.size):
        v = x[i]
        if nan_mask[i]:
            continue
        if v < min_val:
            j = i
            min_val = v
    return j


@ngjit
def nanargmax_jit(x):
    # Get argmax. If empty give -1, if all nan, give -2. This differs from
    # numpy which throws an error.
    x = _atleast_1d(x)
    if x.size == 0:
        return -1
    nan_mask = np.isnan(x)
    if nan_mask.all():
        return -2
    x = x.ravel()
    nan_mask = nan_mask.ravel()
    j = 0
    max_val = -np.inf
    for i in range(x.size):
        v = x[i]
        if nan_mask[i]:
            continue
        if v > max_val:
            j = i
            max_val = v
    return j



import dask.array as da
import numba as nb
import numpy as np
import pytest
import xarray as xr

from raster_tools.distance import proximity as prx
from raster_tools.dtypes import F32
from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster


@pytest.mark.parametrize(
    "x1,x2,truth",
    [
        ([0, 0], [0, 0], 0),
        ([0, 0], [0, 1], 360),
        ([0, 0], [1, 1], 45),
        ([0, 0], [2, 0], 90),
        ([0, 0], [1, -1], 135),
        ([0, 0], [0, -1], 180),
        ([0, 0], [-1, -1], 225),
        ([0, 0], [-1, 0], 270),
        ([0, 0], [-1, 1], 315),
    ],
)
def test_proximity_direction(x1, x2, truth):
    assert prx._direction(*x1, *x2) == truth


esri_src = np.array(
    [
        [0, 1, 1, 0, 0, 0],
        [0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [2, 0, 0, 0, 0, 0],
    ]
)
esri_euc_prox = np.array(
    [
        [1.0, 0.0, 0.0, 1.0, 2.0, 3.0],
        [1.4142135, 1.0, 0.0, 1.0, 2.0, 3.0],
        [2.236068, 1.4142135, 1.0, 1.4142135, 2.236068, 3.1622777],
        [2.0, 2.236068, 2.0, 2.236068, 2.828427, 3.6055512],
        [1.0, 1.4142135, 2.236068, 3.1622777, 3.6055512, 4.2426405],
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    ],
    dtype=np.float32,
)
esri_euc_alloc = np.array(
    [
        [1, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1],
        [2, 1, 1, 1, 1, 1],
        [2, 2, 2, 1, 1, 1],
        [2, 2, 2, 2, 2, 1],
    ]
)
esri_euc_dir = np.array(
    [
        [90.0, 0.0, 0.0, 270.0, 270.0, 270.0],
        [45.0, 360.0, 0.0, 270.0, 270.0, 270.0],
        [26.565052, 45.0, 360.0, 315.0, 296.56506, 288.43494],
        [180.0, 26.565052, 360.0, 333.43494, 315.0, 303.69006],
        [180.0, 225.0, 243.43495, 341.56506, 326.30994, 315.0],
        [0.0, 270.0, 270.0, 270.0, 270.0, 323.1301],
    ],
    dtype=np.float32,
)


def test_proximity_analysis_esri_example():
    src = Raster(esri_src).set_null_value(0)

    prox, alloc, dirn = prx.proximity_analysis(src)

    assert np.allclose(prox, esri_euc_prox)
    assert np.allclose(alloc, esri_euc_alloc)
    assert np.allclose(dirn, esri_euc_dir)
    assert prox.dtype == F32
    assert prox.xdata.compute().dtype == F32
    assert alloc.dtype == src.dtype
    assert alloc.xdata.compute().dtype == src.dtype
    assert dirn.dtype == F32
    assert dirn.xdata.compute().dtype == F32

    mask = esri_euc_prox > 3
    prox_truth = esri_euc_prox.copy()
    prox_truth[mask] = prox.null_value
    alloc_truth = esri_euc_alloc.copy()
    alloc_truth[mask] = alloc.null_value
    dir_truth = esri_euc_dir.copy()
    dir_truth[mask] = dirn.null_value

    prox, alloc, dirn = prx.proximity_analysis(src, max_distance=3)

    assert np.allclose(prox, prox_truth)
    assert np.allclose(prox.mask.compute(), mask)
    assert np.allclose(alloc, alloc_truth)
    assert np.allclose(alloc.mask.compute(), mask)
    assert np.allclose(dirn, dir_truth)
    assert np.allclose(dirn.mask.compute(), mask)


@nb.jit(nopython=True, nogil=True)
def euclidean(x1, x2):
    return np.sqrt(np.sum((x1 - x2) ** 2))


@nb.jit(nopython=True, nogil=True)
def taxi(x1, x2):
    return np.sum(np.abs(x1 - x2))


@nb.jit(nopython=True, nogil=True)
def chessboard(x1, x2):
    return np.max(np.abs(x1 - x2))


@nb.jit(nopython=True, nogil=True)
def haversine(x1, x2):
    x1 = np.radians(x1)
    x2 = np.radians(x2)
    dx = x2 - x1
    sindx = np.sin(dx / 2)
    h = sindx[1] ** 2 + (np.cos(x1[1]) * np.cos(x2[1]) * sindx[0] ** 2)
    return 2 * 6371009.0 * np.arcsin(np.sqrt(h))


@nb.jit(nopython=True, nogil=True)
def apply_metric(x, xc, yc, func, max_distance=np.inf, is_cb=False):
    prox = np.full_like(x, np.nan, dtype=F32)
    alloc = np.full_like(x, -99, dtype=x.dtype)
    dirn = np.full_like(x, np.nan, dtype=F32)
    ny, nx = x.shape

    target_values = np.array([v for v in x.ravel() if v != 0])
    target_locs = np.array(
        [
            (i % x.shape[0], i // x.shape[1])
            for i, v in enumerate(x.ravel())
            if v != 0
        ]
    )
    nt = len(target_values)

    for i in range(ny):
        for j in range(nx):
            x1 = np.array([xc[j], yc[i]])
            x2s = np.zeros((nt, 2))
            for k in range(nt):
                ti, tj = target_locs[k]
                x2s[k] = (xc[ti], yc[tj])
            is_target = False
            for k, x2 in enumerate(x2s):
                is_target = (x1[0] == x2[0]) and (x1[1] == x2[1])
                if is_target:
                    prox[i, j] = 0
                    alloc[i, j] = target_values[k]
                    dirn[i, j] = 0
                    break
            if is_target:
                continue
            dists = np.zeros(nt, dtype=np.float32)
            for k in range(nt):
                x2 = x2s[k]
                dists[k] = func(x1, x2)
            if is_cb:
                imin = 0
                for k in range(nt):
                    x2 = x2s[k]
                    if x2[0] == x1[0] or x2[1] == x1[1]:
                        imin = k
            else:
                imin = np.argmin(dists)
            if dists[imin] > max_distance:
                continue
            x2 = x2s[imin]
            prox[i, j] = dists[imin]
            alloc[i, j] = target_values[imin]
            dirn[i, j] = prx._direction(x1[0], x1[1], x2[0], x2[1])
    return prox, alloc, dirn


def build_raster(src):
    x = np.arange(src.shape[1]) * 30
    y = np.arange(src.shape[0] - 1, -1, -1) * 30
    xrs = xr.DataArray(src, dims=("y", "x"), coords=(y, x))
    rs = Raster(xrs).set_null_value(0)
    return rs


src1 = np.zeros((100, 100))
src1[20] = np.arange(1, 101)
src2 = np.zeros((100, 100))
src2[100 // 2, 100 // 2] = 1
src3 = np.zeros((100, 100))
src3[10:15, 20] = 1
src3[80, 60:65] = np.arange(2, 7)
src3[-1, 0] = 7
srcs = [src1, src2, src2]
srcs = [build_raster(s) for s in srcs]


@pytest.mark.parametrize("max_distance", [np.inf, 20 * 30])
@pytest.mark.parametrize(
    "metric,truth_func",
    [
        ("euclidean", euclidean),
        (None, euclidean),
        ("taxi", taxi),
        ("taxicab", taxi),
        ("manhatten", taxi),
        ("city_block", taxi),
        ("chebyshev", chessboard),
        ("chessboard", chessboard),
    ],
)
@pytest.mark.parametrize("src", srcs)
def test_proximity_metric(src, metric, truth_func, max_distance):
    data = src.values[0]
    x = src.x
    y = src.y

    rprox, ralloc, rdirn = prx.proximity_analysis(
        src, distance_metric=metric, max_distance=max_distance
    )
    tprox, talloc, tdirn = apply_metric(
        data,
        x,
        y,
        truth_func,
        max_distance,
        is_cb=metric in ("chebyshev", "chessboard"),
    )
    mask = np.isnan(tprox)
    tprox[mask] = get_default_null_value(F32)
    talloc[mask] = get_default_null_value(src.dtype)
    tdirn[mask] = get_default_null_value(F32)

    assert np.allclose(rprox, tprox)
    assert rprox.dtype == F32
    assert np.allclose(ralloc, talloc)
    assert ralloc.dtype == src.dtype
    assert np.allclose(rdirn, tdirn)
    assert rdirn.dtype == F32


def test_proximity_great_circle():
    tprox = Raster("tests/data/raster/prox_haversine_proximity.tif")
    talloc = Raster("tests/data/raster/prox_haversine_allocation.tif")
    tdirn = Raster("tests/data/raster/prox_haversine_direction.tif")

    src = Raster("tests/data/raster/prox_src.tif")

    rprox, ralloc, rdirn = prx.proximity_analysis(
        src, distance_metric="haversine"
    )
    assert np.allclose(rprox, tprox)
    assert np.allclose(ralloc, talloc)
    assert np.allclose(rdirn, tdirn)

    max_dist = 1.5e6
    mask = tprox.values > max_dist
    tprox.xdata.data = da.where(mask, tprox.null_value, tprox.data)
    talloc.xdata.data = da.where(mask, talloc.null_value, talloc.data)
    tdirn.xdata.data = da.where(mask, tdirn.null_value, tdirn.data)

    rprox, ralloc, rdirn = prx.proximity_analysis(
        src, distance_metric="haversine", max_distance=max_dist
    )
    assert np.allclose(rprox, tprox)
    assert np.allclose(ralloc, talloc)
    assert np.allclose(rdirn, tdirn)



import os

# Force the use of shapely 2 instead of pygeos in geopandas
os.environ["USE_PYGEOS"] = "0"

import geopandas as gpd  # noqa: E402

# Check if geopandas has already been imported before raster_tools and turn off
# use of pygeos if it is turned on. shapely is required as the geopandas
# backend for line_stats.
if gpd.options.use_pygeos:
    gpd.options.use_pygeos = False



import os
import sys

import dask
import dask.array as da
import dask.dataframe as dd
import dask_geopandas as dgpd
import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio as rio
import xarray as xr
from dask.delayed import delayed
from packaging import version
from rasterio.enums import MergeAlg
from rasterio.env import GDALVersion
from rasterio.features import rasterize as rio_rasterize

from raster_tools.dtypes import (
    F64,
    I8,
    I16,
    I64,
    U64,
    is_float,
    is_int,
    is_str,
)
from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster, get_raster

PYOGRIO_SUPPORTED = sys.version_info >= (3, 8)
if PYOGRIO_SUPPORTED:
    import warnings

    import pyogrio as ogr

    warnings.filterwarnings(
        "ignore",
        message=".*Measured \\(M\\) geometry types are not supported.*",
    )
else:
    import fiona

__all__ = [
    "Vector",
    "get_vector",
    "list_layers",
    "count_layer_features",
    "open_vectors",
]


class VectorError(Exception):
    pass


def _is_series(geo):
    return isinstance(geo, (gpd.GeoSeries, dgpd.GeoSeries))


def _is_frame(geo):
    return isinstance(geo, (gpd.GeoDataFrame, dgpd.GeoDataFrame))


def list_layers(path):
    """List the layers in a vector source file."""
    if not os.path.exists(path):
        raise IOError(f"No such file or directory: {path!r}")
    if PYOGRIO_SUPPORTED:
        layers = ogr.list_layers(path)
        return list(layers[:, 0])
    else:
        return fiona.listlayers(path)


def count_layer_features(path, layer):
    """Count the number of features in a layer of a vector source file.

    Parameters
    ----------
    path : str
        The path to a vector source file.
    layer : str, int
        The layer to count features from. Can be the 0-based index or a string.

    Returns
    -------
    int
        The number of features.

    """
    if not os.path.exists(path):
        raise IOError(f"No such file or directory: {path!r}")
    if not is_int(layer) and not is_str(layer):
        raise TypeError("The layer must be an int or string.")
    if is_int(layer) and layer < 0:
        raise ValueError("The layer must be positive.")

    if PYOGRIO_SUPPORTED:
        info = ogr.read_info(path, layer=layer)
        return info["features"]
    else:
        with fiona.open(path, layer=layer) as src:
            return len(src)


def _df_asign_index(df, index):
    df.index = index
    return df


def _read_file_delayed(path, layer, npartitions):
    total_size = count_layer_features(path, layer)
    batch_size = (total_size // npartitions) + 1

    row_offset = 0
    dfs = []
    divs = []
    while row_offset < total_size:
        left = row_offset
        right = min(row_offset + batch_size, total_size)
        rows = slice(left, right)
        df = delayed(gpd.read_file)(path, rows=rows, layer=layer)
        index = pd.RangeIndex(left, right)
        df = delayed(_df_asign_index)(df, index)
        dfs.append(df)
        divs.append(left)
        row_offset += batch_size
    divs.append(right - 1)
    meta = gpd.read_file(path, layer=layer, rows=1)
    divs = tuple(divs)
    # Returns a dask_geopandas GeoDataFrame
    return dd.from_delayed(dfs, meta, divisions=divs), total_size


def _normalize_layers_arg(layers):
    if isinstance(layers, (list, tuple)):
        if len(layers) == 0:
            raise ValueError("Layers argument was empty")
        first = layers[0]
        is_ftype = is_str if is_str(first) else is_int
        if not all(is_ftype(layer) for layer in layers):
            raise TypeError(
                "Layers must be all name strings or all integer indices"
            )
        return layers
    elif is_int(layers) or is_str(layers):
        return [layers]
    else:
        raise TypeError(f"Could not understand layers argument: {layers!r}")


_TARGET_CHUNK_SIZE = 50_000


def open_vectors(path, layers=None):
    """Open a vector source.

    This can be any GDAL/OGR compatible source.

    Parameters
    ----------
    path : str, Path
        The vector source path.
    layers : list of int, int, list of str, str, or None, optional
        The layer(s) to read from the source. This can be a single layer
        int or string, a list of layer ints/strings, or ``None``. If ``None``,
        all layers are read in. If int(s) are given, they are treated as
        0-based layer indices.

    Returns
    -------
    Vector, list of Vector
        If only a single layer was specified or found, a single Vector object
        is returned. Otherwise, a list of Vector objects is returned.

    """
    if not os.path.exists(path):
        raise IOError(f"No such file or directory: {path!r}")

    src_layers = list_layers(path)
    n_layers = len(src_layers)
    if layers is not None:
        layers = _normalize_layers_arg(layers)
        src_layers_set = set(src_layers)
        if is_str(layers[0]) and any(
            layer not in src_layers_set for layer in layers
        ):
            raise ValueError("Invalid layer name")
        elif is_int(layers[0]) and any(
            layer >= n_layers or layer < 0 for layer in layers
        ):
            raise ValueError(
                "Invalid layer index. Index either outside valid range or less"
                " than 0"
            )
    else:
        layers = src_layers

    dfs = []
    for layer in layers:
        n = count_layer_features(path, layer=layer)
        if PYOGRIO_SUPPORTED:
            df = dgpd.read_file(
                path, layer=layer, chunksize=_TARGET_CHUNK_SIZE
            )
            pair = (df, n)
        else:
            n_parts = max(n // _TARGET_CHUNK_SIZE, 1)
            pair = _read_file_delayed(path, layer, n_parts)
        dfs.append(pair)

    if len(dfs) > 1:
        return [Vector(df, n) for df, n in dfs]
    return Vector(*dfs[0])


def get_vector(src):
    if isinstance(src, Vector):
        return src
    elif is_str(src):
        result = open_vectors(src)
        if not isinstance(result, Vector):
            # More than one layer was found
            raise ValueError("Input source must only have 1 layer")
        return result
    elif _is_series(src) or _is_frame(src):
        return Vector(src)
    else:
        raise TypeError("Invalid vector input")


def _get_len_from_divisions(divs):
    # This should never overcount but may undercount by 1 because
    # dask.dataframe special cases the last division when creating a dataframe.
    #
    # NOTE: Dask dataframes have a divisions property. This is a tuple of
    # locations along the index where partitions start and stop. It looks like
    # this: (0, 10, 20, 29) for a dataframe with 3 divisions and 30 elements.
    # The first element is the start of the first partition and is inclusive.
    # The next is the exclusive end of the first partition and the inclusive
    # start of the second partition. This continues until the last partition
    # where a monkey wrench is thrown in. The last division in the tuple is
    # inclusive but dask treats it like every other division, which makes
    # things less clear and potentially makes the size of the dataframe
    # inexact. Because partitions can be merged etc., there is no way to tell
    # if the last partition in a dataframe you have been given is actually the
    # original last partition. Dask also does not make an effort to keep track
    # of the total length. The best way I have found to get the length without
    # computing it, is to go by the divisions. Because of the special case and
    # the inability to differentiate it from other partitions, this func may
    # undercount by 1. It will give the exact value only if the original last
    # partition is no longer part of given dataframe. It is safer to undercount
    # than overcount however.
    n = 0
    for i in range(len(divs) - 1):
        left = divs[i]
        right = divs[i + 1]
        n += right - left
    return n


def _guess_geo_len(geo):
    if not dask.is_dask_collection(geo):
        return len(geo)

    if geo.known_divisions:
        return _get_len_from_divisions(geo.divisions)
    raise ValueError("Divisions must be set.")


def _get_best_effort_geo_len(geo):
    if not dask.is_dask_collection(geo) or geo.npartitions == 1:
        return len(geo)

    n = _guess_geo_len(geo.partitions[:-1])
    # Evaluate the last partition to get its exact size. Hopefully this is the
    # special-cased partition. If it isn't, this will undercount by 1. See
    # the note in _get_len_from_divisions.
    n += len(geo.partitions[-1])
    return n


_RIO_64BIT_INTS_SUPPORTED = GDALVersion.runtime().at_least("3.5") and (
    version.parse(rio.__version__) >= version.parse("1.3")
)


def _get_rio_dtype(dtype):
    if dtype == I8:
        return I16
    # GDAL >= 3.5 and Rasterio >= 1.3 support 64-bit (u)ints
    if dtype in (I64, U64) and not _RIO_64BIT_INTS_SUPPORTED:
        return F64
    return dtype


def _rio_rasterize_wrapper(
    shape, transform, geometry, values, out_dtype, fill, all_touched
):
    rio_dtype = _get_rio_dtype(out_dtype)
    values_dtype = _get_rio_dtype(values.dtype)
    if values_dtype != values.dtype:
        values = values.astype(values_dtype)

    rast_array = rio_rasterize(
        zip(geometry.values, values),
        out_shape=shape,
        transform=transform,
        fill=fill,
        all_touched=all_touched,
        merge_alg=MergeAlg.replace,
        dtype=rio_dtype,
    )

    if rio_dtype != out_dtype:
        rast_array = rast_array.astype(out_dtype)
    return rast_array


def _rasterize_block(
    xc, yc, geometry, values, out_dtype, fill, all_touched, block_info=None
):
    shape_2d = block_info[None]["chunk-shape"]
    valid = ~(geometry.is_empty | geometry.isna())
    geometry = geometry[valid]
    values = values[valid.values]
    if len(geometry) == 0:
        return np.full(shape_2d, fill, dtype=out_dtype)

    xc = xc.ravel()
    yc = yc.ravel()
    dummy_data = da.zeros(shape_2d)
    dummy_xrs = xr.DataArray(dummy_data, coords=(yc, xc), dims=("y", "x"))
    transform = dummy_xrs.rio.transform()
    return _rio_rasterize_wrapper(
        shape_2d, transform, geometry, values, out_dtype, fill, all_touched
    )


def _rasterize_partition(
    df, xlike, field, fill, target_dtype, all_touched=True
):
    if xlike.ndim > 2:
        # Trim off band dim since its not needed
        xlike = xlike[0]
    geometry = df.geometry
    values = df.index if field is None else df[field]
    values = (
        values.to_dask_array()
        if dask.is_dask_collection(values)
        else values.to_numpy()
    )
    if field is None:
        # Convert to 1-based index
        values = values + 1
        # Convert to minimum dtype
        values = values.astype(target_dtype)
    # Chunk up coords to match data. Dask will provide the correct coord chunks
    # to the rasterizing func in map_blocks
    xc = da.from_array(xlike.x.data, chunks=xlike.chunks[1]).reshape((1, -1))
    yc = da.from_array(xlike.y.data, chunks=xlike.chunks[0]).reshape((-1, 1))

    result_data = da.map_blocks(
        _rasterize_block,
        xc,
        yc,
        chunks=xlike.chunks,
        meta=np.array((), dtype=target_dtype),
        # func args
        geometry=geometry,
        values=values,
        out_dtype=target_dtype,
        fill=fill,
        all_touched=all_touched,
    )
    return result_data


def _vector_to_raster_dask(df, size, xlike, field=None, all_touched=True):
    if not dask.is_dask_collection(xlike):
        raise ValueError("xlike must be a dask collection")

    # There is no way to neatly align the dataframe partitions with the raster
    # chunks. Because of this, we burn in each partition on its own raster and
    # then merge the results.
    fill = 0
    if field is None:
        target_dtype = np.min_scalar_type(size)
    else:
        target_dtype = df[field].dtype
        if target_dtype.kind == "u":
            # Null values are difficult for unsigned ints so cast up to the
            # next largest width signed int. If target_dtype is U64, this will
            # cast to a F64.
            target_dtype = np.promote_types(target_dtype, I16)
        fill = get_default_null_value(target_dtype)
    results = []
    parts = df.partitions if dask.is_dask_collection(df) else [df]
    for part in parts:
        res = _rasterize_partition(
            part,
            xlike,
            field,
            fill,
            target_dtype=target_dtype,
            all_touched=all_touched,
        )
        results.append(res)
    # (n-partitions, yn, xn)
    result_data = da.stack(results, axis=0)
    # (1, yn, xn)
    result_data = result_data.max(axis=0, keepdims=True)
    x = xlike.x.data
    y = xlike.y.data
    result = xr.DataArray(result_data, coords=([1], y, x), dims=xlike.dims)
    if xlike.rio.crs is not None:
        result = result.rio.write_crs(xlike.rio.crs)
    result = result.rio.write_nodata(fill)
    return result


def _geoms_to_raster_mask(xc, yc, geoms, all_touched=True, block_info=None):
    xc = xc.ravel()
    yc = yc.ravel()
    # Convert geoms into a boolean (0/1) array using xc and yc for gridding.
    shape = (yc.size, xc.size)
    transform = xr.DataArray(
        da.zeros(shape), coords=(yc, xc), dims=("y", "x")
    ).rio.transform()
    # Use a MemoryFile to avoid writing to disk
    with rio.io.MemoryFile() as memfile, rio.open(
        memfile,
        mode="w+",
        driver="GTiff",
        width=xc.size,
        height=yc.size,
        count=1,
        crs=geoms.crs,
        transform=transform,
        dtype="uint8",
    ) as ds:
        ds.write(np.ones(shape, dtype="uint8"), 1)
        mask, _ = rio.mask.mask(ds, geoms.values, all_touched=all_touched)
        return mask.squeeze()


def _vector_to_raster_mask(geoms, like, all_touched=True):
    xc, yc = like.get_chunked_coords()
    xc = xc[0]
    yc = yc[0]
    parts = []
    for part in geoms.partitions:
        data = da.map_blocks(
            _geoms_to_raster_mask,
            xc,
            yc,
            geoms=part,
            all_touched=all_touched,
            chunks=like.data.chunks[1:],
            meta=np.array((), dtype=I8),
        )
        parts.append(data)
    data = da.stack(parts, axis=0).max(axis=0, keepdims=True)
    x = like.x
    y = like.y
    xrs = xr.DataArray(data, coords=([1], y, x), dims=like.xdata.dims)
    if like.crs is not None:
        xrs = xrs.rio.write_crs(like.crs)
    xrs = xrs.rio.write_nodata(0)
    return xrs


def _normalize_geo_data(geo):
    if not _is_series(geo) and not _is_frame(geo):
        raise TypeError(
            "Invalid data type. Must be some type GeoDataFrame or"
            f" GeoSeries. Got {type(geo)}."
        )
    if _is_series(geo):
        geo = geo.to_frame("geometry")
    if dask.is_dask_collection(geo) and not geo.known_divisions:
        raise ValueError(
            "Unknown divisions set on input data. Divisions must be set."
        )
    return geo


_GMTRY = "geometry"


class Vector:
    """A class representing vector data.

    Take care to provide the actual number of features in the data when
    creating a new Vector.
    """

    def __init__(self, geo, size=None):
        """Create a `Vector` object.

        Parameters
        ----------
        geo : GeoDataFrame or GeoSeries
            The vector data to use. This can be a geopandas or dask_geopandas
            object.
        size : int, optional
            The number of features. If not provided, A best effort will be made
            to determine the number of features. Part of the data may be
            temporarily loaded.

        """
        geo = _normalize_geo_data(geo)
        if not dask.is_dask_collection(geo):
            self._size = len(geo)
            geo = dgpd.from_geopandas(geo, npartitions=1)
            self._geo = geo
        else:
            self._geo = geo
            # NOTE: self._size is not data-dependant so when adding features
            # that can affect length, care must be taken to properly update
            # _size.
            if size is None:
                self._size = _get_best_effort_geo_len(geo)
            else:
                if not is_int(size):
                    raise TypeError("Size must be an int.")

                # The guess will underestimate by 1 in most cases
                size_guess = _guess_geo_len(geo)
                if size == size_guess or size == (size_guess + 1):
                    self._size = size
                else:
                    raise ValueError(
                        "Given size does not match divisions in data"
                    )

    def __len__(self):
        return self._size

    def __repr__(self):
        lazy = dask.is_dask_collection(self._geo)
        return "<Vector(shapes: {}, fields: {}, CRS: {}, lazy:{})>".format(
            self.size, self.shape[1], self.crs, lazy
        )

    @property
    def table(self):
        """The vector attribute table."""
        return self._geo.drop(columns=[_GMTRY])

    @property
    def data(self):
        return self._geo

    @property
    def size(self):
        """
        The number of vectors contained. NaN is returned if the vector is lazy.

        Use `len(vector)` to trigger computation of the length.
        """
        return len(self)

    @property
    def shape(self):
        """The shape of the attribute table."""
        if not dask.is_dask_collection(self._geo):
            return self.table.shape
        return (self.size, len(self.field_names))

    @property
    def crs(self):
        """The vector CRS."""
        return rio.crs.CRS.from_wkt(self._geo.crs.to_wkt())

    @property
    def field_schema(self):
        """A mapping of the field names to dtypes."""
        return self.table.dtypes.to_dict()

    @property
    def field_names(self):
        """The field names."""
        return list(self.field_schema.keys())

    @property
    def field_dtypes(self):
        """The field dtypes."""
        return list(self.field_schema.values())

    @property
    def geometry(self):
        """The vector geometry series."""
        return self._geo.geometry

    @property
    def tasks(self):
        """The number of lazy operations left to be computed."""
        if dask.is_dask_collection(self._geo):
            # TODO: take into account partitions?
            return len(self._geo.dask)
        return 0

    @property
    def bounds(self):
        """Return a bounds array or dask array: (minx, miny, maxx, maxy).

        If the vector is lazy, the output is a dask array.
        """
        return self.geometry.total_bounds

    def copy(self):
        """Copies the vector."""
        return Vector(self._geo.copy())

    def eval(self):
        """Computes the built-up chain of operations on the underlying data."""
        if dask.is_dask_collection(self._geo):
            return Vector(self._geo.compute(), self._size)
        return self

    def to_lazy(self, npartitions=None):
        """Converts to a lazy dask-backed Vector.

        Parameters
        ----------
        npartitions : int
            The number of partitions to use.

        Returns
        -------
        Vector
            A lazy copy of the vector.

        """
        if dask.is_dask_collection(self._geo):
            return self
        if npartitions is None:
            # TODO: calculate better value based on df size
            npartitions = 1
        return Vector(
            dgpd.from_geopandas(self._geo, npartitions=npartitions), self.size
        )

    def to_dask(self):
        """
        Returns the underlying data as a :obj:`dask_geopandas.GeoDataFrame`.
        """
        return self.to_lazy()._geo

    def to_dataframe(self):
        """Returns the underlying GeoDataFrame."""
        return self.data

    def to_shapely(self):
        """Returns the vector data as a list of shapely geometries objects."""
        return self.geometry.to_list()

    def to_raster(self, like, field=None, all_touched=True):
        """Convert vector data to a raster.

        Parameters
        ----------
        like : Raster
            A to use for grid and CRS information. The resulting raster will be
            on the same grid as `like`.
        field : str, optional
            The name of a field to use for fill values when rasterizing the
            vector features.
        all_touched : bool, optional
            If ``True``, grid cells that the vector touches will be burned in.
            If False, only cells with a center point inside of the vector
            perimeter will be burned in.

        Returns
        -------
        Raster
            The resulting raster. The result will have a single band. Each
            vector shape is assigned a 1-based integer value equal to its order
            in the original vector collection. This integer value is what is
            burned in at the corresponding grid cells. The dtype of the result
            will be the minimum, unsigned integer type that can fit the ID
            values. The null value is 0.

        """
        like = get_raster(like)
        if field is not None:
            if not is_str(field):
                raise TypeError("Field must be a string")
            if field not in self._geo:
                raise ValueError(f"Invalid field name: {repr(field)}")
            dtype = self.field_schema[field]
            if not is_int(dtype) and not is_float(dtype):
                raise ValueError(
                    "The specified field must be a scalar data type"
                )

        xrs = _vector_to_raster_dask(
            self.to_crs(like.crs)._geo,
            self.size,
            xlike=like.xdata,
            field=field,
            all_touched=all_touched,
        )
        return Raster(xrs)

    def to_raster_mask(self, like, all_touched=True):
        """Convert vector data to a raster mask.

        Parameters
        ----------
        like : Raster
            A to use for grid and CRS information. The resulting raster will be
            on the same grid as `like`.
        all_touched : bool, optional
            If ``True``, grid cells that the vector touches will be burned in.
            If False, only cells with a center point inside of the vector
            perimeter will be burned in.

        Returns
        -------
        Raster
            The resulting raster. The result will have a single band. All cells
            that fall under the vector data are masked with ``1``. The null
            value is 0.

        """
        like = get_raster(like)

        xrs = _vector_to_raster_mask(
            self.to_crs(like.crs).geometry, like, all_touched=all_touched
        )
        return Raster(xrs)

    def to_crs(self, crs):
        """Transform the vector coordinates to the specified `crs`.

        Parameters
        ----------
        crs : str, :obj:`pyproj.CRS`
            The crs to transform the vector coordinates to. The value can be
            anything accepted by
            :meth:`pyproj.CRS.from_user_input()`,
            such as an EPSG string (eg "EPSG:4326") or a WKT string.

        Returns
        -------
        The transformed vector(s).

        """
        return Vector(self._geo.to_crs(crs))

    def cast_field(self, name, dtype):
        """Cast a field to the specified `dtype`.

        Parameters
        ----------
        name : str
            The field name to cast.
        dtype : str, type, or numpy.dtype
            The dtype to cast to.

        Returns
        -------
        Vector
            A new vector with the updated attribute table.

        """
        if not is_str(name):
            raise TypeError("Name must be a string")
        if name not in self._geo:
            raise ValueError(f"Inalid attribute name: {name}")
        geo = self._geo.copy()
        geo[name] = geo[name].astype(dtype)
        return Vector(geo)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise NotImplementedError("Slicing is not supported")
        if not is_int(idx):
            raise TypeError(f"Invalid index type: {type(idx)}")
        if idx >= self.size:
            raise IndexError(
                f"Index {idx} is too large for vector with size {self.size}"
            )
        # Convert negative indices to positive because the index only handles +
        if idx < 0:
            old_idx = idx
            idx = self.size + idx
            if idx < 0:
                raise IndexError(
                    f"Negative index {old_idx} is too small for vector with "
                    f"size {self.size}."
                )
        # [[idx]] forces loc to return a data frame. The new frames index is
        # [idx], however. This breaks future iteration we reset it to [0] to
        # allow iteration over the resulting Vector.
        subgeo = self._geo.loc[[idx]]
        subgeo.index = dask.array.from_array([0]).to_dask_dataframe()
        return Vector(subgeo, 1)

    def buffer(self, *args, **kwargs):
        """Apply the buffer operation to the vector geometries.

        Parameters
        ----------
        *args : positional args
            Arguments to be passed on to GoePandas.
        **kwargs : keyword args
            Keyword arguments to be passed onto GeoPandas.

        Returns
        -------
        Vector
            The buffered vector.

        """
        return Vector(self._geo.buffer(*args, **kwargs))

    def simplify(self, *args, **kwargs):
        """Simplify the vector geometries.

        Parameters
        ----------
        *args : positional args
            Arguments to be passed on to GoePandas.
        **kwargs : keyword args
            Keyword arguments to be passed onto GeoPandas.

        Returns
        -------
        Vector
            The simplify vector.

        """
        return Vector(self._geo.simplify(*args, **kwargs))

    def save(self, path, **fiona_kwargs):
        """Save the vector data to a file.

        By default, an ESRI shapefile is written.

        Parameters
        ----------
        path : str
            Output location to save the vector data to.
        **fiona_kwargs : keyword args
            Keyword arguments to pass on to :func:`fiona.open` such as
            ``layer='name'``, ``driver='GPKG'``, or ``mode='a'``

        Returns
        -------
        Vector
            Returns a new vector pointing to the saved location.

        """
        (geo,) = dask.compute(self._geo)
        geo.to_file(path, **fiona_kwargs)
        return open_vectors(path)
